"""Generate the C++ functions to iterate over instances."""

import io
from typing import (
    Optional,
    Dict,
    List,
    Tuple,
    Sequence,
    Set,
    Final,
    FrozenSet,
)

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Error,
    Identifier,
    assert_never,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.cpp import (
    common as cpp_common,
    naming as cpp_naming,
    yielding as cpp_yielding,
)
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)
from aas_core_codegen.intermediate import construction as intermediate_construction
from aas_core_codegen.yielding import flow as yielding_flow


# region Check


@ensure(lambda result: not (result is not None) or (len(result) >= 1))
def _verify_that_property_enum_literals_do_not_collide(
    symbol_table: intermediate.SymbolTable,
) -> Optional[List[Error]]:
    """Check that the literal names for the properties do not collied within a class."""
    errors = []  # type: List[Error]

    # NOTE (mristin, 2023-10-07):
    # We use getter name as string representation for the enum ``Property``, so we have
    # to make sure that there are no conflicts.
    literal_name_to_getter_and_prop = (
        dict()
    )  # type: Dict[str, Tuple[str, intermediate.Property]]

    for cls in symbol_table.classes:
        observed_literal_names = dict()  # type: Dict[str, str]
        for prop in cls.properties:
            literal_name = cpp_naming.enum_literal_name(prop.name)

            conflicting_property_name = observed_literal_names.get(literal_name, None)

            if conflicting_property_name is not None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The property {prop.name!r} and "
                        f"the property {conflicting_property_name!r} conflict in "
                        f"the C++ Property literal name {literal_name!r} "
                        f"in class {cls.name!r}",
                    )
                )
                continue

            getter = cpp_naming.getter_name(prop.name)

            another_getter_and_prop = literal_name_to_getter_and_prop.get(
                literal_name, None
            )
            if another_getter_and_prop is not None:
                another_getter, another_prop = another_getter_and_prop

                if another_getter != getter:
                    errors.append(
                        Error(
                            cls.parsed.node,
                            f"The property {prop.name!r} from class {cls.name!r} and "
                            f"the property {another_prop.name!r} "
                            f"from class {another_prop.specified_for.name!r} "
                            f"have differing getter names, "
                            f"{getter!r} and {another_getter!r}, respectively, "
                            f"for the literal in C++ enum Property {literal_name!r}",
                        )
                    )
                    continue
            else:
                literal_name_to_getter_and_prop[literal_name] = (getter, prop)

            observed_literal_names[literal_name] = prop.name

    if len(errors) > 0:
        return errors

    return None


# endregion

# region Generation


def _generate_property_enum(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the enum which represents all the properties of all the classes."""
    literal_name_set = set()  # type: Set[Identifier]
    for cls in symbol_table.classes:
        for prop in cls.properties:
            literal_name_set.add(cpp_naming.enum_literal_name(prop.name))

    literal_names = sorted(literal_name_set)

    literal_definitions = [
        f"{literal_name} = {i}" for i, literal_name in enumerate(literal_names)
    ]

    literal_definitions_joined = ",\n".join(literal_definitions)

    property_enum = cpp_naming.enum_name(Identifier("Property"))

    return Stripped(
        f"""\
/**
 * Define the properties over all the classes to compactly represent the paths.
 */
enum class {property_enum} : std::uint32_t {{
{I}{indent_but_first_line(literal_definitions_joined, I)}
}};"""
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the C++ header code of the iteration functions and structures."""
    collision_errors = _verify_that_property_enum_literals_do_not_collide(
        symbol_table=symbol_table
    )
    if collision_errors is not None:
        return None, collision_errors

    namespace = Stripped(f"{library_namespace}::iteration")

    include_guard_var = cpp_common.include_guard_var(namespace)

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    property_enum = cpp_naming.enum_name(Identifier("Property"))
    property_to_wstring = cpp_naming.function_name(Identifier("property_to_wstring"))

    blocks = [
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/types.hpp"

#pragma warning(push, 0)
#include <deque>
#include <iterator>
#include <memory>
#include <string>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(library_namespace),
        Stripped(
            """\
/**
 * \\defgroup iteration Define functions and structures to iterate over instances.
 * @{
*/
namespace iteration {"""
        ),
        Stripped("// region Pathing"),
        _generate_property_enum(symbol_table=symbol_table),
        Stripped(
            f"""\
std::wstring {property_to_wstring}(
{I}{property_enum} property
);"""
        ),
        Stripped(
            f"""\
/**
 * Represent a segment of a path to some value.
 */
class ISegment {{
 public:
{I}virtual std::wstring ToWstring() const = 0;
{I}virtual std::unique_ptr<ISegment> Clone() const = 0;
{I}virtual ~ISegment() = default;
}};  // class ISegment"""
        ),
        Stripped(
            f"""\
/**
 * Represent a property access on a path.
 */
struct PropertySegment : public ISegment {{
{I}/**
{I} * Enumeration of the property
{I} */
{I}Property property;

{I}PropertySegment(
{II}Property a_property
{I});

{I}std::wstring ToWstring() const override;
{I}std::unique_ptr<ISegment> Clone() const override;

{I}~PropertySegment() override = default;
}};  // struct PropertySegment"""
        ),
        Stripped(
            f"""\
/**
 * Represent an index access on a path.
 */
struct IndexSegment : public ISegment {{
{I}/**
{I} * Index of the item
{I} */
{I}size_t index;

{I}explicit IndexSegment(
{II}size_t an_index
{I});

{I}std::wstring ToWstring() const override;
{I}std::unique_ptr<ISegment> Clone() const override;

{I}~IndexSegment() override = default;
}};  // struct IndexSegment"""
        ),
        Stripped(
            f"""\
/**
 * \\brief Represent a path to some value.
 *
 * This is a path akin to C++ expressions. It is not to be confused with different
 * paths used in the specification. This path class is meant to help with reporting.
 * For example, we can use this path to let the user know when there is
 * a verification error in a model which can concern instances, but also properties
 * and items in the lists.
 */
struct Path {{
{I}// NOTE (mristin):
{I}// We did not implement the reflection at the moment since we did not have a use
{I}// case for it. If you need reflection, please contact the developers. It should
{I}// be a small step going from paths to dereferencing to getters and setters.

{I}std::deque<std::unique_ptr<ISegment> > segments;

{I}Path();
{I}Path(const Path& other);
{I}Path(Path&& other);
{I}Path& operator=(const Path& other);
{I}Path& operator=(Path&& other);

{I}std::wstring ToWstring() const;
}};  // struct Path"""
        ),
        Stripped("// endregion Pathing"),
        Stripped("// region Iterators and descent"),
        Stripped(
            f"""\
/// \\cond HIDDEN
namespace impl {{
class IIterator {{
 public:
{I}virtual void Start() = 0;
{I}virtual void Next() = 0;
{I}virtual bool Done() const = 0;
{I}virtual const std::shared_ptr<types::IClass>& Get() const = 0;
{I}virtual long Index() const = 0;

{I}/// Prepend the segments to the path reflecting where this iterator points to.
{I}virtual void PrependToPath(Path* path) const = 0;

{I}virtual std::unique_ptr<IIterator> Clone() const = 0;

{I}virtual ~IIterator() = default;
}};  // class IIterator
}}  // namespace impl
/// \\endcond"""
        ),
        Stripped(
            f"""\
/**
 * \\brief Iterate over an AAS instance.
 *
 * Unlike STL, this is <em>not</em> a light-weight iterator. We implement
 * a "yielding" iterator by leveraging code generation so that we always keep
 * the model stack as well as the properties iterated thus far.
 *
 * This means that copy-construction and equality comparisons are much more heavy-weight
 * than you'd usually expect from an STL iterator. For example, if you want to sort
 * model instances, you are most probably faster if you populate a vector, and then
 * sort the vector.
 *
 * Also, given that this iterator is not light-weight, you should in almost all cases
 * avoid the postfix increment (it++) and prefer the prefix one (++it) as the postfix
 * increment would create an iterator copy every time.
 *
 * The value of the iterator is intentionally constant reference to a shared pointer.
 * This merely means that you can not change the <em>pointer</em> while you are
 * iterating. The pointed instances, however, is freely mutable. This way you can make
 * further shared pointers, or include the pointed instances in other collections
 * different from the original container. On the other hand, the normal case, where
 * the pointer is only de-referenced, remains efficient as no copy of
 * the shared pointer is created.
 *
 * We follow the C++ standard, and assume that comparison between the two iterators
 * over two different instances results in undefined behavior. See
 * http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2009/n2948.html and
 * https://stackoverflow.com/questions/4657513/comparing-iterators-from-different-containers.
 *
 * Since we use const references to shared pointers here you can also share ownership
 * over instances in your own external containers. Making a copy of a shared pointer
 * will automatically increase reference count, even though there is a constant
 * reference. Since we do not make copies of the shared pointers, it is very important
 * that the given shared pointers outlive the iteration, lest cause undefined behavior.
 *
 * Changing the references <em>during</em> the iteration invalidates the iterators and
 * results in undefined behavior. This is similar to many of the containers in the STL,
 * see: https://stackoverflow.com/questions/6438086/iterator-invalidation-rules-for-c-containers
 *
 * See these StackOverflow questions for performance related to shared pointers and
 * constant references to shared pointers (copying <em>versus</em> referencing):
 * * https://stackoverflow.com/questions/12002480/passing-stdshared-ptr-to-constructors/12002668#12002668
 * * https://stackoverflow.com/questions/3310737/should-we-pass-a-shared-ptr-by-reference-or-by-value
 * * https://stackoverflow.com/questions/37610494/passing-const-shared-ptrt-versus-just-shared-ptrt-as-parameter
 *
 * The following StackOverflow question and answers go into more detail how const-ness
 * and shared pointers fit together:
 * https://stackoverflow.com/questions/36271663/why-does-copying-a-const-shared-ptr-not-violate-const-ness
 */
class Iterator {{
{I}using iterator_category = std::forward_iterator_tag;
{I}/// The difference is meaningless, but has to be defined.
{I}using difference_type = std::ptrdiff_t;
{I}using value_type = std::shared_ptr<types::IClass>;
{I}using pointer = const std::shared_ptr<types::IClass>*;
{I}using reference = const std::shared_ptr<types::IClass>&;

 public:
{I}Iterator(const Iterator& other);
{I}Iterator(Iterator&& other);

{I}Iterator& operator=(const Iterator& other);
{I}Iterator& operator=(Iterator&& other);

{I}reference operator*() const;
{I}pointer operator->();

{I}// Prefix increment
{I}Iterator& operator++();

{I}// Postfix increment
{I}Iterator operator++(int);

{I}friend bool operator==(const Iterator& a, const Iterator& b);
{I}friend bool operator!=(const Iterator& a, const Iterator& b);

{I}friend class Descent;
{I}friend class DescentOnce;
{I}friend Path MaterializePath(const Iterator& iterator);
{I}friend void PrependToPath(const Iterator& iterator, Path* path);

 private:
{I}explicit Iterator(
{II}std::unique_ptr<impl::IIterator> implementation
{I}) :
{II}implementation_(std::move(implementation)) {{
{II}  // Intentionally empty.
{I}}}

{I}std::unique_ptr<impl::IIterator> implementation_;
}};"""
        ),
        Stripped("bool operator==(const Iterator& a, const Iterator& b);"),
        Stripped("bool operator!=(const Iterator& a, const Iterator& b);"),
        Stripped(
            """\
/**
 * \\brief Materialize the path that the \\p iterator points to.
 *
 * We assume that you always want a copy of the path, rather than inspect
 * the path during the iteration.
 *
 * \\param iterator for which we want to materialize the path
 * \\return Path referring to the pointed instance
 */
Path MaterializePath(const Iterator& iterator);"""
        ),
        Stripped(
            f"""\
/**
 * Build a facade over an instance to iterate over instances referenced from it.
 */
class IDescent {{
 public:
{I}virtual Iterator begin() const = 0;
{I}virtual const Iterator& end() const = 0;
{I}virtual ~IDescent() = default;
}};  // class IDescent"""
        ),
        Stripped(
            f"""\
/**
 * \\brief Provide a recursive iterable over all the instances referenced from
 * an instance.
 *
 * Please see the notes in the class Iterator regarding the constant reference to
 * a shared pointer. In short, the instance should outlive the descent, so make
 * sure you do not destroy it during the descent.
 *
 * Range-based loops should fit the vast majority of the use cases:
 * \\code
 * std::shared_ptr<types::Environment> env = ...;
 * for (
 * {I}const std::shared_ptr<types::IClass>& instance
 * {I}: Descent(env)
 * ) {{
 * {I}do_something(instance);
 * }}
 * \\endcode
 *
 * \\param that instance to be iterated over recursively
 * \\return Iterable over referenced instances
 */
class Descent : public IDescent {{
 public:
{I}Descent(
{II}std::shared_ptr<types::IClass> instance
{I});

{I}Iterator begin() const override;
{I}const Iterator& end() const override;

{I}~Descent() override = default;

 private:
{I}std::shared_ptr<types::IClass> instance_;
}};  // class Descent"""
        ),
        Stripped(
            f"""\
/**
 * \\brief Provide a non-recursive iterable over the instances referenced from
 * an instance.
 *
 * Please see the notes in the class Iterator regarding the constant reference to
 * a shared pointer. In short, the instance should outlive the descent, so make
 * sure you do not destroy it during the descent.
 *
 * Range-based loops should fit the vast majority of the use cases:
 * \\code
 * std::shared_ptr<types::Environment> env = ...;
 * for (
 * {I}const std::shared_ptr<types::IClass>& instance
 * {I}: DescentOnce(env)
 * ) {{
 * {I}do_something(instance);
 * }}
 * \\endcode
 */
class DescentOnce : public IDescent {{
 public:
{I}DescentOnce(
{II}std::shared_ptr<types::IClass> instance
{I});

{I}Iterator begin() const override;
{I}const Iterator& end() const override;

{I}~DescentOnce() override = default;

 private:
{I}std::shared_ptr<types::IClass> instance_;
}};  // class DescentOnce"""
        ),
        Stripped("// endregion Iterators and descent"),
    ]  # type: List[Stripped]

    if len(symbol_table.enumerations) > 0:
        blocks.append(Stripped("// region Over enumerations"))

        for enum in symbol_table.enumerations:
            enum_name = cpp_naming.enum_name(enum.name)
            over_enum = cpp_naming.constant_name(Identifier(f"over_{enum.name}"))

            blocks.append(
                Stripped(
                    f"""\
/**
 * \\brief Give a container for all the literals of types::{enum_name}.
 *
 * This container is practical when you want to show the literals in a GUI or a CLI.
 */
extern const std::vector<types::{enum_name}> {over_enum};"""
                )
            )

        blocks.append(Stripped("// endregion Over enumerations"))

    blocks.extend(
        [
            Stripped(
                """\
}  // namespace iteration
/**@}*/"""
            ),
            cpp_common.generate_namespace_closing(library_namespace),
            cpp_common.WARNING,
            Stripped(f"#endif  // {include_guard_var}"),
        ]
    )

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue(), None


def _generate_property_to_wstring_implementation(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the implementation of the stringification for ``Property`` enum."""
    literal_name_to_getter = dict()  # type: Dict[Identifier, Identifier]
    literal_name_set = set()  # type: Set[Identifier]

    for cls in symbol_table.classes:
        for prop in cls.properties:
            literal_name = cpp_naming.enum_literal_name(prop.name)
            literal_name_to_getter[literal_name] = cpp_naming.getter_name(prop.name)

            literal_name_set.add(literal_name)

    literal_names = sorted(literal_name_set)

    property_enum = cpp_naming.enum_name(Identifier("Property"))

    case_blocks = []  # type: List[Stripped]
    for literal_name in literal_names:
        getter = literal_name_to_getter[literal_name]

        case_blocks.append(
            Stripped(
                f"""\
case {property_enum}::{literal_name}:
{I}return {cpp_common.wstring_literal(getter)};"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}throw std::invalid_argument(
{II}common::Concat(
{III}"Unexpected property literal: ",
{III}std::to_string(
{IIII}static_cast<std::uint32_t>(property)
{III})
{II})
{I});"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    property_to_wstring = cpp_naming.function_name(Identifier("property_to_wstring"))

    return Stripped(
        f"""\
/**
 * Translate the enumeration literal \\p property to text.
 *
 * \\param property to be converted into text
 * \\return text representation of \\p property
 * \\throw std::invalid_argument if \\p property invalid
 */
std::wstring {property_to_wstring}(
{I}Property property
) {{
{I}switch (property) {{
{II}{indent_but_first_line(case_blocks_joined, II)}
{I}}}
}}  // function to_wstring"""
    )


def _generate_property_segment_implementation() -> List[Stripped]:
    """Generate the implementation of ``PropertySegment`` struct."""
    property_to_wstring = cpp_naming.function_name(Identifier("property_to_wstring"))

    return [
        Stripped("// region struct PropertySegment"),
        Stripped(
            f"""\
PropertySegment::PropertySegment(Property a_property) {{
{I}property = a_property;
}}"""
        ),
        Stripped(
            f"""\
std::wstring PropertySegment::ToWstring() const {{
{I}return common::Concat(
{II}L".",
{II}{property_to_wstring}(property)
{I});
}}"""
        ),
        Stripped(
            f"""\
std::unique_ptr<ISegment> PropertySegment::Clone() const {{
{I}return common::make_unique<PropertySegment>(*this);
}}"""
        ),
        Stripped("// endregion struct PropertySegment"),
    ]


def _generate_index_segment_implementation() -> List[Stripped]:
    """Generate the implementation of ``IndexSegment`` struct."""
    return [
        Stripped("// region struct IndexSegment"),
        Stripped(
            f"""\
IndexSegment::IndexSegment(size_t an_index) {{
{I}index = an_index;
}}"""
        ),
        Stripped(
            f"""\
std::wstring IndexSegment::ToWstring() const {{
{I}return common::Concat(
{II}L"[",
{II}std::to_wstring(index),
{II}L"]"
{I});
}}"""
        ),
        Stripped(
            f"""\
std::unique_ptr<ISegment> IndexSegment::Clone() const {{
{I}return common::make_unique<IndexSegment>(*this);
}}"""
        ),
        Stripped("// endregion struct IndexSegment"),
    ]


def _generate_path_implementation() -> List[Stripped]:
    """Generate the implementation of the ``Path`` struct."""
    return [
        Stripped("// region struct Path"),
        Stripped(
            f"""\
Path::Path() {{
{I}// Intentionally empty.
}}"""
        ),
        Stripped(
            f"""\
Path::Path(const Path& other) {{
{I}for (const std::unique_ptr<ISegment>& segment : other.segments) {{
{II}segments.emplace_back(segment->Clone());
{I}}}
}}"""
        ),
        Stripped(
            f"""\
Path::Path(Path&& other) {{
{I}segments = std::move(other.segments);
}}"""
        ),
        Stripped(
            f"""\
Path& Path::operator=(const Path& other) {{
{I}segments.clear();
{I}for (const std::unique_ptr<ISegment>& segment : other.segments) {{
{II}segments.emplace_back(segment->Clone());
{I}}}
{I}return *this;
}}"""
        ),
        Stripped(
            f"""\
Path& Path::operator=(Path&& other) {{
{I}if (this != &other) {{
{II}segments = std::move(other.segments);
{I}}}
{I}return *this;
}}"""
        ),
        Stripped(
            f"""\
std::wstring Path::ToWstring() const {{
{I}std::vector<std::wstring> parts;
{I}parts.reserve(segments.size());

{I}for (const std::unique_ptr<ISegment>& segment : segments ) {{
{II}parts.emplace_back(segment->ToWstring());
{I}}}

{I}size_t size = 0;
{I}for (const std::wstring& part : parts) {{
{II}size += part.size();
{I}}}

{I}std::wstring result;
{I}result.reserve(size);
{I}for (const std::wstring& part : parts) {{
{II}result.append(part);
{I}}}

{I}return result;
}}"""
        ),
        Stripped("// endregion struct Path"),
    ]


class IteratorQualities:
    """Query the qualities of a non-recursive iterator corresponding to a class."""

    #: The class corresponding to the iterator
    cls: Final[intermediate.ConcreteClass]

    #: The properties which should be iterated over
    relevant_properties: Final[Sequence[intermediate.Property]]

    #: A set of Python IDs of the relevant properties
    relevant_property_id_set: Final[FrozenSet[int]]

    #: Set if the class contains a property which is a list of instances
    cls_contains_a_list_property: Final[bool]

    def __init__(self, cls: intermediate.ConcreteClass) -> None:
        """Initialize with the given class."""
        relevant_properties = []  # type: List[intermediate.Property]

        cls_contains_a_list_property = False

        for prop in cls.properties:
            type_anno = intermediate.beneath_optional(prop.type_annotation)

            if isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
                type_anno.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ):
                relevant_properties.append(prop)

            elif isinstance(type_anno, intermediate.ListTypeAnnotation):
                assert isinstance(
                    type_anno.items, intermediate.OurTypeAnnotation
                ) and isinstance(
                    type_anno.items.our_type,
                    (intermediate.AbstractClass, intermediate.ConcreteClass),
                ), (
                    f"NOTE (mristin, 2023-09-27): We expect only lists of classes "
                    f"at the moment, but you specified {prop.type_annotation} "
                    f"in class {cls.name!r} and property {prop.name!r}. "
                    f"Please contact the developers if you need this feature."
                )

                cls_contains_a_list_property = True

                relevant_properties.append(prop)
            else:
                pass

        self.cls = cls
        self.relevant_properties = relevant_properties
        self.relevant_property_id_set = frozenset(
            id(prop) for prop in self.relevant_properties
        )
        self.cls_contains_a_list_property = cls_contains_a_list_property


@require(lambda iterator_qualities: len(iterator_qualities.relevant_properties) == 0)
def _generate_empty_iterator_over_cls(
    iterator_qualities: IteratorQualities,
) -> List[Stripped]:
    """Generate the iterator over a class with no references to other instances."""
    interface_name = cpp_naming.interface_name(iterator_qualities.cls.name)
    iterator_over_cls = cpp_naming.class_name(
        Identifier(f"Iterator_over_{iterator_qualities.cls.name}")
    )

    return [
        Stripped(
            f"""\
/**
 * This iterator is always done as {interface_name}
 * references no other instances.
 */
class {iterator_over_cls} : public impl::IIterator {{
 public:
{I}{iterator_over_cls}(
{II}const std::shared_ptr<types::IClass>&
{I}) {{
{II}// Intentionally empty.
{I}}}

{I}void Start() override {{
{II}// Intentionally empty.
{I}}}

{I}void Next() override {{
{II}throw std::logic_error(
{III}"You want to move "
{III}"an {iterator_over_cls}, "
{III}"but the iterator is always done as "
{III}"{interface_name} "
{III}"references no other instances."
{II});
{I}}}

{I}bool Done() const override {{
{II}return true;
{I}}}

{I}const std::shared_ptr<types::IClass>& Get() const override {{
{II}throw std::logic_error(
{III}"You want to get from an {iterator_over_cls}, "
{III}"but the iterator is always done as "
{III}"{interface_name} references "
{III}"no other instances."
{II});
{I}}}

{I}long Index() const override {{
{II}return -1;
{I}}}

{I}std::unique_ptr<impl::IIterator> Clone() const override {{
{II}return common::make_unique<{iterator_over_cls}>(*this);
{I}}}

{I}void PrependToPath(Path*) const override {{
{II}throw std::logic_error(
{III}"You want to prepend to path from an {iterator_over_cls}, "
{III}"but the iterator is always done as "
{III}"{interface_name} references "
{III}"no other instances."
{II});
{I}}}

{I}~{iterator_over_cls}() override = default;
}};  // class {iterator_over_cls}"""
        )
    ]


@require(lambda iterator_qualities: len(iterator_qualities.relevant_properties) > 0)
def _generate_iterator_over_cls_execute_implementation(
    iterator_qualities: IteratorQualities,
) -> Stripped:
    """Generate the implementation of ``Execute()`` member."""
    cls = iterator_qualities.cls

    flow = [
        yielding_flow.command_from_text(
            """\
property_.reset();
item_ = nullptr;
index_ = -1;
done_ = false;"""
        )
    ]  # type: List[yielding_flow.Node]

    if iterator_qualities.cls_contains_a_list_property:
        flow.append(yielding_flow.command_from_text("cursor_.reset();"))

    for prop in iterator_qualities.relevant_properties:
        type_anno = intermediate.beneath_optional(prop.type_annotation)

        getter_name = cpp_naming.getter_name(prop.name)
        property_literal = cpp_naming.enum_literal_name(prop.name)

        if isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
            type_anno.our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
                flow.append(
                    yielding_flow.IfTrue(
                        f"casted_->{getter_name}().has_value()",
                        [
                            yielding_flow.command_from_text(
                                f"""\
property_ = Property::{property_literal};
item_ = std::move(
{I}std::static_pointer_cast<types::IClass>(
{II}*(casted_->{getter_name}())
{I})
);
++index_;"""
                            ),
                            yielding_flow.Yield(),
                        ],
                    )
                )
            else:
                flow.append(
                    yielding_flow.command_from_text(
                        f"""\
property_ = Property::{property_literal};
item_ = std::move(
{I}std::static_pointer_cast<types::IClass>(
{II}casted_->{getter_name}()
{I})
);
++index_;"""
                    )
                )
                flow.append(yielding_flow.Yield())

        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            assert isinstance(
                type_anno.items, intermediate.OurTypeAnnotation
            ) and isinstance(
                type_anno.items.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ), (
                f"NOTE (mristin, 2023-09-27): We expect only lists of classes "
                f"at the moment, but you specified {prop.type_annotation} "
                f"in class {cls.name!r} and property {prop.name!r}. "
                f"Please contact the developers if you need this feature."
            )

            if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
                list_type = cpp_common.generate_type_with_const_ref_if_applicable(
                    type_annotation=prop.type_annotation.value,
                    types_namespace=cpp_common.TYPES_NAMESPACE,
                )
                list_var = cpp_naming.variable_name(Identifier(f"the_{prop.name}"))

                flow.append(
                    yielding_flow.IfTrue(
                        f"casted_->{getter_name}().has_value()",
                        [
                            yielding_flow.command_from_text(
                                f"property_ = Property::{property_literal};"
                            ),
                            yielding_flow.For(
                                f"*cursor_ < casted_->{getter_name}()->size()",
                                "++(*cursor_);",
                                [
                                    yielding_flow.command_from_text(
                                        f"""\
{list_type} {list_var}(
{I}*(casted_->{getter_name}())
);

item_ = std::move(
{I}std::static_pointer_cast<types::IClass>(
{II}{list_var}[*cursor_]
{I})
);
++index_;"""
                                    ),
                                    yielding_flow.Yield(),
                                ],
                                init="cursor_ = 0;",
                            ),
                            yielding_flow.command_from_text("cursor_.reset();"),
                        ],
                    )
                )
            else:
                list_type = cpp_common.generate_type_with_const_ref_if_applicable(
                    type_annotation=prop.type_annotation,
                    types_namespace=cpp_common.TYPES_NAMESPACE,
                )
                list_var = cpp_naming.variable_name(Identifier(f"the_{prop.name}"))

                flow.append(
                    yielding_flow.command_from_text(
                        f"property_ = Property::{property_literal};"
                    )
                )
                flow.append(
                    yielding_flow.For(
                        f"*cursor_ < casted_->{getter_name}().size()",
                        "++(*cursor_);",
                        [
                            yielding_flow.command_from_text(
                                f"""\
{list_type} {list_var}(
{I}casted_->{getter_name}()
);

item_ = std::move(
{I}std::static_pointer_cast<types::IClass>(
{II}{list_var}[*cursor_]
{I})
);
++index_;"""
                            ),
                            yielding_flow.Yield(),
                        ],
                        init="cursor_ = 0;",
                    )
                )
                flow.append(yielding_flow.command_from_text("cursor_.reset();"))

    flow.append(
        yielding_flow.command_from_text(
            """\
done_ = true;
index_ = -1;"""
        )
    )

    body = cpp_yielding.generate_execute_body(
        flow=flow, state_member=Identifier("state_")
    )

    iterator_name = cpp_naming.class_name(Identifier(f"Iterator_over_{cls.name}"))

    return Stripped(
        f"""\
void {iterator_name}::Execute() {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


@require(lambda iterator_qualities: len(iterator_qualities.relevant_properties) > 0)
def _generate_iterator_over_cls(
    iterator_qualities: IteratorQualities,
) -> List[Stripped]:
    """Generate a non-recursive iterator over referenced instances."""
    cls = iterator_qualities.cls

    iterator_name = cpp_naming.class_name(Identifier(f"Iterator_over_{cls.name}"))

    interface_name = cpp_naming.interface_name(cls.name)

    private_properties = [
        Stripped(
            """\
// We make instance_ a pointer, so that we can follow the rule-of-zero.
const std::shared_ptr<types::IClass>* instance_;"""
        ),
        Stripped(
            f"""\
// We make casted_ a pointer, so that we can follow the rule-of-zero.
const types::{interface_name}* casted_;"""
        ),
        Stripped("std::uint32_t state_;"),
        Stripped("common::optional<Property> property_;"),
    ]
    if iterator_qualities.cls_contains_a_list_property:
        private_properties.append(
            Stripped("common::optional<size_t> cursor_;  // in yield-from loops")
        )

    private_properties.extend(
        (
            Stripped("std::shared_ptr<types::IClass> item_;"),
            Stripped("long index_;  // in total iteration"),
            Stripped("bool done_;"),
        )
    )

    private_properties_joined = "\n".join(private_properties)

    execute_block = _generate_iterator_over_cls_execute_implementation(
        iterator_qualities=iterator_qualities
    )

    if iterator_qualities.cls_contains_a_list_property:
        prepend_to_path_block = Stripped(
            f"""\
void {iterator_name}::PrependToPath(
{I}Path* path
) const {{
{I}#ifdef DEBUG
{I}if (Done()) {{
{II}throw std::logic_error(
{III}"You want to prepend to path from {iterator_name}, "
{III}"but the iterator was done."
{II});
{I}}}
{I}#endif

{I}if (cursor_.has_value()) {{
{II}path->segments.emplace_front(
{III}common::make_unique<IndexSegment>(*cursor_)
{II});
{I}}}

{I}#ifdef DEBUG
{I}if (!property_.has_value()) {{
{II}throw std::logic_error(
{III}"You want to prepend to path from {iterator_name}, "
{III}"but the property_ has not been set to a value."
{II});
{I}}}
{I}#endif

{I}path->segments.emplace_front(
{II}common::make_unique<PropertySegment>(*property_)
{I});
}}"""
        )

    else:
        prepend_to_path_block = Stripped(
            f"""\
void {iterator_name}::PrependToPath(
{I}Path* path
) const {{
{I}#ifdef DEBUG
{I}if (Done()) {{
{II}throw std::logic_error(
{III}"You want to prepend to path from {iterator_name}, "
{III}"but the iterator was done."
{II});
{I}}}

{I}if (!property_.has_value()) {{
{II}throw std::logic_error(
{III}"You want to prepend to path from {iterator_name}, "
{III}"but the property_ has not been set to a value."
{II});
{I}}}
{I}#endif

{I}path->segments.emplace_front(
{II}common::make_unique<PropertySegment>(*property_)
{I});
}}"""
        )

    return [
        Stripped(
            f"""\
/**
 * Iterate non-recursively over the instances referenced from an instance.
 */
class {iterator_name} : public impl::IIterator {{
 public:
{I}{iterator_name}(
{II}const std::shared_ptr<types::IClass>& instance
{I});
{I}void Start() override;
{I}void Next() override;
{I}bool Done() const override;
{I}const std::shared_ptr<types::IClass>& Get() const override;
{I}long Index() const override;
{I}void PrependToPath(Path* path) const override;
{I}std::unique_ptr<impl::IIterator> Clone() const override;
{I}~{iterator_name}() override = default;

 private:
{I}{indent_but_first_line(private_properties_joined, I)}

{I}void Execute();
}};  // class {iterator_name}"""
        ),
        Stripped(
            f"""\
{iterator_name}::{iterator_name}(
{I}const std::shared_ptr<types::IClass>& instance
) :
{I}instance_(&instance),
{I}// NOTE (mristin):
{I}// The dynamic cast is necessary due to virtual inheritance. Otherwise,
{I}// we would have used static cast.
{I}casted_(
{II}dynamic_cast<types::{interface_name}*>(
{III}instance.get()
{II})
{I}) {{
{I}// Intentionally empty.
}}"""
        ),
        Stripped(
            f"""\
void {iterator_name}::Start() {{
{I}state_ = 0;
{I}Execute();
}}"""
        ),
        Stripped(
            f"""\
void {iterator_name}::Next() {{
{I}#ifdef DEBUG
{I}if (Done()) {{
{II}throw std::logic_error(
{III}"You want to move {iterator_name}, "
{III}"but it was done."
{II});
{I}}}
{I}#endif

{I}Execute();
}}"""
        ),
        Stripped(
            f"""\
bool {iterator_name}::Done() const {{
{I}return done_;
}}"""
        ),
        Stripped(
            f"""\
const std::shared_ptr<types::IClass>& {iterator_name}::Get() const {{
{I}#ifdef DEBUG
{I}if (Done()) {{
{II}throw std::logic_error(
{III}"You want to get from {iterator_name}, "
{III}"but the iterator was done."
{II});
{I}}}
{I}#endif

{I}return item_;
}}"""
        ),
        Stripped(
            f"""\
long {iterator_name}::Index() const {{
{I}#ifdef DEBUG
{I}if (Done() && index_ != -1) {{
{II}throw std::logic_error(
{III}common::Concat(
{IIII}"Expected index to be -1 "
{IIII}"from a done {iterator_name}, "
{IIII}"but got: ",
{IIII}std::to_string(index_)
{III})
{II});
{I}}}
{I}#endif

{I}return index_;
}}"""
        ),
        prepend_to_path_block,
        Stripped(
            f"""\
std::unique_ptr<impl::IIterator> {iterator_name}::Clone() const {{
{I}return common::make_unique<{iterator_name}>(*this);
}}"""
        ),
        execute_block,
    ]


def _generate_iteration_over_cls(cls: intermediate.ConcreteClass) -> List[Stripped]:
    """Generate the iterator over the given class."""
    iterator_qualities = IteratorQualities(cls=cls)

    if len(iterator_qualities.relevant_properties) == 0:
        return _generate_empty_iterator_over_cls(iterator_qualities=iterator_qualities)

    interface_name = cpp_naming.interface_name(cls.name)

    return [
        Stripped(f"// region Non-recursive iteration over {interface_name}"),
        *_generate_iterator_over_cls(iterator_qualities=iterator_qualities),
        Stripped("// endregion"),
    ]


def _generate_new_non_recursive_iterator_function(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the factory for non-recursive iterators."""
    case_blocks = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        enum_name = cpp_naming.enum_name(Identifier("Model_type"))
        literal_name = cpp_naming.enum_literal_name(cls.name)

        iterator_over_cls = cpp_naming.class_name(
            Identifier(f"Iterator_over_{cls.name}")
        )

        case_blocks.append(
            Stripped(
                f"""\
case types::{enum_name}::{literal_name}:
{I}return common::make_unique<{iterator_over_cls}>(
{II}instance
{I});"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}throw std::logic_error(
{II}common::Concat(
{III}"Unexpected model type: ",
{III}std::to_string(
{IIII}static_cast<std::uint32_t>(instance->model_type())
{III})
{II})
{I});"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    switch_stmt = Stripped(
        f"""\
switch (instance->model_type()) {{
{I}{indent_but_first_line(case_blocks_joined, I)}
}}"""
    )

    new_non_recursive_iterator = cpp_naming.function_name(
        Identifier("new_non_recursive_iterator")
    )

    return Stripped(
        f"""\
/**
 * Produce a non-recursive iterator over the instance given its runtime model type.
 */
std::unique_ptr<impl::IIterator> {new_non_recursive_iterator}(
{I}const std::shared_ptr<types::IClass>& instance
) {{
{I}{indent_but_first_line(switch_stmt, I)}
}}"""
    )


def _generate_recursive_inclusive_iterator_execute() -> Stripped:
    """Generate the ``Execute()`` method of the recursive inclusive iterator."""
    new_non_recursive_iterator = cpp_naming.function_name(
        Identifier("new_non_recursive_iterator")
    )

    flow = [
        yielding_flow.command_from_text(
            """\
item_ = instance_;
index_ = 0;
done_ = false;
non_recursive_iterator_.reset(nullptr);
recursive_iterator_.reset(nullptr);"""
        ),
        yielding_flow.Yield(),
        yielding_flow.command_from_text(
            f"""\
non_recursive_iterator_ = {new_non_recursive_iterator}(
{I}*instance_
);"""
        ),
        yielding_flow.For(
            "!non_recursive_iterator_->Done()",
            "non_recursive_iterator_->Next();",
            [
                yielding_flow.command_from_text(
                    """\
item_ = &(non_recursive_iterator_->Get());
++index_;"""
                ),
                yielding_flow.Yield(),
                yielding_flow.command_from_text(
                    f"""\
recursive_iterator_ = std::move(
{I}common::make_unique<RecursiveExclusiveIterator>(
{II}*item_
{I})
);"""
                ),
                yielding_flow.For(
                    "!recursive_iterator_->Done()",
                    "recursive_iterator_->Next();",
                    [
                        yielding_flow.command_from_text(
                            """\
item_ = &(recursive_iterator_->Get());
++index_;"""
                        ),
                        yielding_flow.Yield(),
                    ],
                    init="recursive_iterator_->Start();",
                ),
                yielding_flow.command_from_text("recursive_iterator_.reset(nullptr);"),
            ],
            init="non_recursive_iterator_->Start();",
        ),
        yielding_flow.command_from_text(
            """\
non_recursive_iterator_.reset(nullptr);
done_ = true;
index_ = -1;"""
        ),
    ]  # type: List[yielding_flow.Node]

    body = cpp_yielding.generate_execute_body(
        flow=flow, state_member=Identifier("state_")
    )

    return Stripped(
        f"""\
void RecursiveInclusiveIterator::Execute() {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_recursive_iteration() -> List[Stripped]:
    """Generate the iterator to recursively iterate over the referenced instances."""
    execute_block = _generate_recursive_inclusive_iterator_execute()

    return [
        Stripped(
            f"""\
/**
 * Iterate recursively over the instance, including the instance in the iteration.
 *
 * This is a realisation of the following pseudo-code:
 * \\code
 * stack = new Stack();
 * stack.push(instance);
 * while not stack.empty():
 *     instance = stack.pop()
 *     yield instance
 *
 *     it = new_non_recursive_iterator(instance)
 *     while not it.done():
 *         yield recursively from it.get()
 *         it.next()
 * \\endcode
 */
class RecursiveInclusiveIterator : public impl::IIterator {{
 public:
{I}RecursiveInclusiveIterator(
{II}const std::shared_ptr<types::IClass>& instance
{I});

{I}RecursiveInclusiveIterator(
{II}const RecursiveInclusiveIterator& other
{I});
{I}RecursiveInclusiveIterator(
{II}RecursiveInclusiveIterator&& other
{I});
{I}RecursiveInclusiveIterator& operator=(
{II}const RecursiveInclusiveIterator& other
{I});
{I}RecursiveInclusiveIterator& operator=(
{II}RecursiveInclusiveIterator&& other
{I});

{I}void Start() override;
{I}void Next() override;
{I}bool Done() const override;
{I}const std::shared_ptr<types::IClass>& Get() const override;
{I}long Index() const override;
{I}void PrependToPath(Path* path) const override;
{I}std::unique_ptr<impl::IIterator> Clone() const override;
{I}~RecursiveInclusiveIterator() override = default;

 private:
{I}// The instance_ needs to be a pointer so that we can re-assign it in
{I}// the constructors and assignment operations.
{I}const std::shared_ptr<types::IClass>* instance_;

{I}// Iterator over the instances referenced from this instance
{I}// in the outer loop
{I}std::unique_ptr<impl::IIterator> non_recursive_iterator_;

{I}// Iterator for recursion into the reference referenced from this instance
{I}// in the inner loop
{I}std::unique_ptr<impl::IIterator> recursive_iterator_;

{I}const std::shared_ptr<types::IClass>* item_;

{I}bool done_;
{I}long index_;
{I}size_t state_;

{I}void Execute();
}};  // class RecursiveInclusiveIterator"""
        ),
        Stripped(
            f"""\
/**
 * Iterate recursively over the instance, excluding the instance in the iteration.
 *
 * This is a realisation of the following pseudo-code:
 * \\code
 * stack = new Stack();
 * stack.push(instance);
 * while not stack.empty():
 *     some_instance = stack.pop()
 *     if some_instance is not instance:
 *         yield some_instance
 *
 *     it = new_non_recursive_iterator(some_instance)
 *     while not it.done():
 *         yield recursively from it.get()
 *         it.next()
 * \\endcode
 */
class RecursiveExclusiveIterator : public impl::IIterator {{
 public:
{I}RecursiveExclusiveIterator(
{II}const std::shared_ptr<types::IClass>& instance
{I});

{I}void Start() override;
{I}void Next() override;
{I}bool Done() const override;
{I}const std::shared_ptr<types::IClass>& Get() const override;
{I}long Index() const override;
{I}void PrependToPath(Path* path) const override;
{I}std::unique_ptr<impl::IIterator> Clone() const override;
{I}~RecursiveExclusiveIterator() override = default;

 private:
{I}RecursiveInclusiveIterator inclusive_iterator_;
}};  // class RecursiveExclusiveIterator"""
        ),
        Stripped("// region RecursiveInclusiveIterator implementation"),
        Stripped(
            f"""\
RecursiveInclusiveIterator::RecursiveInclusiveIterator(
{I}const std::shared_ptr<types::IClass>& instance
) : instance_(&instance), item_(nullptr), index_(-1) {{
{I}// Intentionally empty.
}}"""
        ),
        Stripped(
            f"""\
RecursiveInclusiveIterator::RecursiveInclusiveIterator(
{I}const RecursiveInclusiveIterator& other
) {{
{I}instance_ = other.instance_;
{I}non_recursive_iterator_ = (other.non_recursive_iterator_ == nullptr)
{II}? nullptr
{II}: other.non_recursive_iterator_->Clone();
{I}recursive_iterator_ = (other.recursive_iterator_ == nullptr)
{II}? nullptr
{II}: other.recursive_iterator_->Clone();
{I}item_ = other.item_;
{I}done_ = other.done_;
{I}index_ = other.index_;
{I}state_ = other.state_;
}}"""
        ),
        Stripped(
            f"""\
RecursiveInclusiveIterator::RecursiveInclusiveIterator(
{I}RecursiveInclusiveIterator&& other
) {{
{I}instance_ = other.instance_;
{I}non_recursive_iterator_ = std::move(other.non_recursive_iterator_);
{I}recursive_iterator_ = std::move(other.recursive_iterator_);
{I}item_ = other.item_;
{I}done_ = other.done_;
{I}index_ = other.index_;
{I}state_ = other.state_;
}}"""
        ),
        Stripped(
            f"""\
RecursiveInclusiveIterator& RecursiveInclusiveIterator::operator=(
{I}const RecursiveInclusiveIterator& other
) {{
{I}return *this = RecursiveInclusiveIterator(other);
}}"""
        ),
        Stripped(
            f"""\
RecursiveInclusiveIterator& RecursiveInclusiveIterator::operator=(
{I}RecursiveInclusiveIterator&& other
) {{
{I}if (this != &other) {{
{II}instance_ = other.instance_;
{II}non_recursive_iterator_ = std::move(other.non_recursive_iterator_);
{II}recursive_iterator_ = std::move(other.recursive_iterator_);
{II}item_ = other.item_;
{II}done_ = other.done_;
{II}index_ = other.index_;
{II}state_ = other.state_;
{I}}}

{I}return *this;
}}"""
        ),
        Stripped(
            f"""\
void RecursiveInclusiveIterator::Start() {{
{I}state_ = 0;
{I}Execute();

{I}#ifdef DEBUG
{I}if (Done()) {{
{II}throw std::logic_error(
{III}"Expected RecursiveInclusiveIterator not to be done at start, but it was."
{II});
{I}}}

{I}if (Index() !== 0) {{
{II}throw std::logic_error(
{III}common::Concat(
{IIII}"Expected RecursiveInclusiveIterator::Index() to be 0 on Start()"
{IIII}", but got ",
{IIII}std::to_string(Index())
{III})
{II});
{I}}}

{I}const std::shared_ptr<IClass>& current_item(Get());
{I}if (current_item == nullptr) {{
{II}throw std::logic_error(
{III}"Unexpected null pointer from Get() at the end of "
{III}"RecursiveInclusiveIterator::Start"
{II});
{I}}}

{I}if (current_item.get() != instance_.get()) {{
{II}throw std::logic_error(
{III}common::Concat(
{IIII}"Expected the current item to point to the instance "
{IIII}"at the end of RecursiveInclusiveIterator::Start, "
{IIII}"but got ",
{IIII}std::to_string(current_item.get()),
{IIII}" from Get() instead of ",
{IIII}std::to_string(instance_.get())
{III})
{II});
{I}}}
{I}#endif
}}"""
        ),
        Stripped(
            f"""\
void RecursiveInclusiveIterator::Next() {{
{I}#ifdef DEBUG
{I}if (Done()) {{
{II}throw std::logic_error(
{III}"You want to move a RecursiveInclusiveIterator, but it was done."
{II});
{I}}}
{I}#endif

{I}Execute();
}}"""
        ),
        Stripped(
            f"""\
bool RecursiveInclusiveIterator::Done() const {{
{I}return done_;
}}"""
        ),
        Stripped(
            f"""\
const std::shared_ptr<types::IClass>& RecursiveInclusiveIterator::Get() const {{
{I}#ifdef DEBUG
{I}if (Done()) {{
{II}throw std::logic_error(
{III}"You want to get from RecursiveInclusiveIterator, but it was done."
{II});
{I}}}

{I}if (item_ == nullptr) {{
{II}throw std::logic_error(
{III}"You want to get from a RecursiveInclusiveIterator, "
{III}"but item_ has not been set."
{II});
{I}}}
{I}#endif

{I}return *item_;
}}"""
        ),
        Stripped(
            f"""\
long RecursiveInclusiveIterator::Index() const {{
{I}#ifdef DEBUG
{I}if (Done() && index_ != -1) {{
{II}throw std::logic_error(
{III}common::Concat(
{IIII}"Expected index to be -1 on a done RecursiveInclusiveIterator, "
{IIII}"but got: ",
{IIII}std::to_string(index_)
{III})
{II});
{I}}}
{I}#endif

{I}return index_;
}}"""
        ),
        Stripped(
            f"""\
void RecursiveInclusiveIterator::PrependToPath(Path* path) const {{
{I}#ifdef DEBUG
{I}if (Done()) {{
{II}throw std::logic_error(
{III}"You want to prepend to path from RecursiveInclusiveIterator, "
{III}"but the iterator was done."
{II});
{I}}}
{I}#endif

{I}if (Index() == 0) {{
{II}// Index set to 0 indicates that the iterator points to the instance itself.
{II}// Therefore, there is nothing to prepend to the path.
{II}return;
{I}}}

{I}if (recursive_iterator_ != nullptr) {{
{II}recursive_iterator_->PrependToPath(path);
{I}}}

{I}if (non_recursive_iterator_ != nullptr) {{
{II}non_recursive_iterator_->PrependToPath(path);
{I}}}
}}"""
        ),
        Stripped(
            f"""\
std::unique_ptr<impl::IIterator> RecursiveInclusiveIterator::Clone() const {{
{I}return common::make_unique<RecursiveInclusiveIterator>(*this);
}}"""
        ),
        execute_block,
        Stripped("// endregion RecursiveInclusiveIterator implementation"),
        Stripped("// region RecursiveExclusiveIterator implementation"),
        Stripped(
            f"""\
RecursiveExclusiveIterator::RecursiveExclusiveIterator(
{I}const std::shared_ptr<types::IClass>& instance
) : inclusive_iterator_(instance) {{
{I}// Intentionally empty.
}}"""
        ),
        Stripped(
            f"""\
void RecursiveExclusiveIterator::Start() {{
{I}inclusive_iterator_.Start();

{I}#ifdef DEBUG
{I}if (inclusive_iterator_.Done()) {{
{II}throw std::logic_error(
{III}"Expected the inclusive iterator to be not-done immediately after start, "
{III}"as the first item is expected to point to the instance itself, "
{III}"but the inclusive iterator was done."
{II});
{I}}}
{I}#endif

{I}// Simply skip the instance in the very first yield.
{I}inclusive_iterator_.Next();
}}"""
        ),
        Stripped(
            f"""\
void RecursiveExclusiveIterator::Next() {{
{I}#ifdef DEBUG
{I}if (Done()) {{
{II}throw std::logic_error(
{III}"You want to move a RecursiveExclusiveIterator, but it was done."
{II});
{I}}}
{I}#endif

{I}inclusive_iterator_.Next();
}}"""
        ),
        Stripped(
            f"""\
bool RecursiveExclusiveIterator::Done() const {{
{I}return inclusive_iterator_.Done();
}}"""
        ),
        Stripped(
            f"""\
const std::shared_ptr<types::IClass>& RecursiveExclusiveIterator::Get() const {{
{I}#ifdef DEBUG
{I}if (Done()) {{
{II}throw std::logic_error(
{III}"You want to get from RecursiveExclusiveIterator, but it was done."
{II});
{I}}}
{I}#endif

{I}return inclusive_iterator_.Get();
}}"""
        ),
        Stripped(
            f"""\
long RecursiveExclusiveIterator::Index() const {{
{I}if (inclusive_iterator_.Done()) {{
{II}return -1;
{I}}}

{I}return inclusive_iterator_.Index() - 1;
}}"""
        ),
        Stripped(
            f"""\
void RecursiveExclusiveIterator::PrependToPath(Path* path) const {{
{I}inclusive_iterator_.PrependToPath(path);
}}"""
        ),
        Stripped(
            f"""\
std::unique_ptr<impl::IIterator> RecursiveExclusiveIterator::Clone() const {{
{I}return common::make_unique<RecursiveExclusiveIterator>(*this);
}}"""
        ),
        Stripped("// endregion RecursiveExclusiveIterator implementation"),
    ]


def _generate_descent_and_descent_once_implementations() -> List[Stripped]:
    """Generate the implementation of the descent iterables."""
    new_non_recursive_iterator = cpp_naming.function_name(
        Identifier("new_non_recursive_iterator")
    )

    return [
        Stripped(
            f"""\
// region Descent

// NOTE (mristin):
// We have to make a copy of the pointer since we would lose otherwise
// in range-based `for` loops,
// see: https://stackoverflow.com/questions/29990045/temporary-lifetime-in-range-for-expression
Descent::Descent(
{I}std::shared_ptr<types::IClass> instance
) : instance_(std::move(instance)) {{
{I}// Intentionally empty.
}}

Iterator Descent::begin() const {{
{I}std::unique_ptr<impl::IIterator> it_impl(
{II}std::move(
{III}common::make_unique<RecursiveExclusiveIterator>(instance_)
{II})
{I});

{I}it_impl->Start();

{I}// NOTE(mristin):
{I}// We short-circuit here for memory frugality,
{I}// as we can immediately dispose it_impl.
{I}if (it_impl->Done()) {{
{II}return end();
{I}}}

{I}return Iterator(std::move(it_impl));
}}

const Iterator& Descent::end() const {{
{I}static Iterator iterator(common::make_unique<AlwaysDoneIterator>());
{I}return iterator;
}}

// endregion Descent"""
        ),
        Stripped(
            f"""\
// region DescentOnce

// NOTE (mristin):
// We have to make a copy of the pointer since we would lose otherwise
// in range-based `for` loops,
// see: https://stackoverflow.com/questions/29990045/temporary-lifetime-in-range-for-expression
DescentOnce::DescentOnce(
{I}std::shared_ptr<types::IClass> instance
) : instance_(std::move(instance)) {{
{I}// Intentionally empty.
}}

Iterator DescentOnce::begin() const {{
{I}std::unique_ptr<impl::IIterator> it_impl(
{II}{new_non_recursive_iterator}(instance_)
{I});

{I}it_impl->Start();

{I}// NOTE(mristin):
{I}// We short-circuit here for efficiency, as we can immediately dispose it_impl.
{I}if (it_impl->Done()) {{
{II}return Iterator(std::move(common::make_unique<AlwaysDoneIterator>()));
{I}}}

{I}return Iterator(std::move(it_impl));
}}

const Iterator& DescentOnce::end() const {{
{I}static Iterator iterator(common::make_unique<AlwaysDoneIterator>());
{I}return iterator;
}}

// endregion DescentOnce"""
        ),
    ]


def _generate_constructor_implementation(cls: intermediate.ConcreteClass) -> Stripped:
    """Transpile the constructor implementation for the given ``cls``."""
    assert all(
        isinstance(stmt, intermediate_construction.AssignArgument)
        for stmt in cls.constructor.inlined_statements
    ), (
        f"We expect only assigns in the inlined constructors, "
        f"but got for class {cls.name!r}: {cls.constructor.inlined_statements}"
    )

    body_statements = []  # type: List[Stripped]

    for stmt in cls.constructor.inlined_statements:
        assert isinstance(stmt, intermediate_construction.AssignArgument), (
            f"Only assigns expected in inlined constructors, but got the following "
            f"statement for the class {cls.name!r}: {stmt}"
        )

        prop_member = cpp_naming.private_property_name(stmt.name)
        arg_name = cpp_naming.argument_name(stmt.argument)

        if stmt.default is not None:
            if isinstance(stmt.default, intermediate_construction.EmptyList):
                body_statements.append(
                    Stripped(
                        f"""\
if ({arg_name}.has_value()) {{
{I}{prop_member} = *{arg_name};
}} else {{
{I}{prop_member}.emplace();
}}"""
                    )
                )
            elif isinstance(stmt.default, intermediate_construction.DefaultEnumLiteral):
                enum_name = cpp_naming.enum_name(stmt.default.enum.name)
                literal_name = cpp_naming.enum_literal_name(stmt.default.literal.name)

                body_statements.append(
                    Stripped(
                        f"""\
{prop_member} = ({arg_name}.has_value())
{I}? *{arg_name}
{I}: {enum_name}::{literal_name};"""
                    )
                )
            else:
                assert_never(stmt.default)

        else:
            prop = cls.properties_by_name[stmt.name]
            if cpp_common.is_referencable(prop.type_annotation):
                body_statements.append(
                    Stripped(f"{prop_member} = std::move({arg_name});")
                )
            else:
                body_statements.append(Stripped(f"{prop_member} = {arg_name};"))

    body = (
        "\n\n".join(body_statements)
        if len(body_statements) > 0
        else "// Intentionally empty."
    )

    cls_name = cpp_naming.class_name(cls.name)

    if len(cls.constructor.arguments) == 0:
        return Stripped(
            f"""\
{cls_name}::{cls_name}() {{
{I}{indent_but_first_line(body, I)}
}}"""
        )

    constructor_argument_specs = []  # type: List[str]
    for arg in cls.constructor.arguments:
        arg_type = cpp_common.generate_type(arg.type_annotation)
        arg_name = cpp_naming.argument_name(arg.name)

        constructor_argument_specs.append(f"{arg_type} {arg_name}")

    constructor_arguments_specs_joined = ",\n".join(constructor_argument_specs)

    return Stripped(
        f"""\
{cls_name}::{cls_name}(
{I}{indent_but_first_line(constructor_arguments_specs_joined, I)}
) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_always_done_iterator() -> Stripped:
    """Generate the definition and implementation of the end-of-descent iterator."""
    return Stripped(
        f"""\
/**
 * This iterator is always done.
 *
 * It is used for efficient comparisons against end-of-descent.
 */
class AlwaysDoneIterator : public impl::IIterator {{
 public:
{I}void Start() override {{
{II}// Intentionally empty.
{I}}}

{I}void Next() override {{
{II}throw std::logic_error(
{III}"You want to move an AlwaysDoneIterator, "
{III}"but the iterator is always done, as its name suggests."
{II});
{I}}}

{I}bool Done() const override {{
{II}return true;
{I}}}

{I}const std::shared_ptr<types::IClass>& Get() const override {{
{II}throw std::logic_error(
{III}"You want to get from an AlwaysDoneIterator, "
{III}"but the iterator is always done, as its name suggests."
{II});
{I}}}

{I}std::unique_ptr<IIterator> Clone() const override {{
{II}return common::make_unique<AlwaysDoneIterator>(*this);
{I}}};

{I}void PrependToPath(Path*) const override {{
{II}throw std::logic_error(
{III}"You want to prepend to path from an AlwaysDoneIterator, "
{III}"but the iterator is always done, as its name suggests."
{II});
{I}}}

{I}long Index() const override {{
{II}return -1;
{I}}}

{I}~AlwaysDoneIterator() override = default;
}};  // class AlwaysDoneIterator"""
    )


def _generate_iterator_implementation() -> List[Stripped]:
    """Generate the impl. of the facade ``Iterator`` around ``impl::Iterator``."""
    return [
        Stripped(
            f"""\
Iterator::Iterator(
{I}const Iterator& other
) : implementation_(other.implementation_->Clone()) {{
{I}// Intentionally empty.
}}"""
        ),
        Stripped(
            f"""\
Iterator::Iterator(
{I}Iterator&& other
) : implementation_(std::move(other.implementation_)) {{
{I}// Intentionally empty.
}}"""
        ),
        Stripped(
            f"""\
Iterator& Iterator::operator=(const Iterator& other) {{
{I}return *this = Iterator(other);
}}"""
        ),
        Stripped(
            f"""\
Iterator& Iterator::operator=(Iterator&& other) {{
{I}if (this != &other) {{
{II}this->implementation_ = std::move(other.implementation_);
{I}}}

{I}return *this;
}}"""
        ),
        Stripped(
            f"""\
const std::shared_ptr<types::IClass>& Iterator::operator*() const {{
{I}if (implementation_->Done()) {{
{II}throw std::logic_error(
{III}"You want to dereference a completed iterator."
{II});
{I}}}

{I}return implementation_->Get();
}}"""
        ),
        Stripped(
            f"""\
const std::shared_ptr<types::IClass>* Iterator::operator->() {{
{I}if (implementation_->Done()) {{
{II}throw std::logic_error(
{III}"You want to dereference a completed iterator."
{II});
{I}}}

{I}return &(implementation_->Get());
}}"""
        ),
        Stripped(
            f"""\
// Prefix increment
Iterator& Iterator::operator++() {{
{I}if (implementation_->Done()) {{
{II}throw std::logic_error(
{III}"You want to move a completed iterator."
{II});
{I}}}

{I}implementation_->Next();
{I}return *this;
}}"""
        ),
        Stripped(
            f"""\
// Postfix increment
Iterator Iterator::operator++(int) {{
{I}Iterator result(*this);
{I}++(*this);
{I}return result;
}}"""
        ),
        Stripped(
            f"""\
bool operator==(const Iterator& a, const Iterator& b) {{
{I}return a.implementation_->Index() == b.implementation_->Index();
}}"""
        ),
        Stripped(
            f"""\
bool operator!=(const Iterator& a, const Iterator& b) {{
{I}return a.implementation_->Index() != b.implementation_->Index();
}}"""
        ),
        Stripped(
            f"""\
Path MaterializePath(const Iterator& iterator) {{
{I}if (iterator.implementation_->Done()) {{
{II}throw std::logic_error(
{III}"You want to materialize path of a completed iterator."
{II});
{I}}}

{I}Path path;
{I}iterator.implementation_->PrependToPath(&path);
{I}return path;
}}"""
        ),
        Stripped(
            f"""\
void PrependToPath(const Iterator& iterator, Path* path) {{
{I}if (iterator.implementation_->Done()) {{
{II}throw std::logic_error(
{III}"You want to prepend a path of a completed iterator."
{II});
{I}}}

{I}iterator.implementation_->PrependToPath(path);
}}"""
        ),
    ]


def _generate_over_enum_implementation(enum: intermediate.Enumeration) -> Stripped:
    """Generate the implementation for a container over ``enum`` literals."""
    enum_name = cpp_naming.enum_name(enum.name)
    over_enum = cpp_naming.constant_name(Identifier(f"over_{enum.name}"))

    literals = []  # type: List[Stripped]
    for literal in enum.literals:
        literal_name = cpp_naming.enum_literal_name(literal.name)
        literals.append(Stripped(f"types::{enum_name}::{literal_name}"))

    literals_joined = ",\n".join(literals)

    return Stripped(
        f"""\
const std::vector<types::{enum_name}> {over_enum} = {{
{I}{indent_but_first_line(literals_joined, I)}
}};"""
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the C++ implementation of the iteration functions and structures."""
    namespace = Stripped(f"{library_namespace}::iteration")

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f'''\
#include "{include_prefix_path}/common.hpp"
#include "{include_prefix_path}/iteration.hpp"'''
        ),
        cpp_common.generate_namespace_opening(namespace),
        Stripped("// region Pathing"),
        _generate_property_to_wstring_implementation(symbol_table=symbol_table),
        *_generate_property_segment_implementation(),
        *_generate_index_segment_implementation(),
        *_generate_path_implementation(),
        Stripped("// endregion Pathing"),
        Stripped("// region Non-recursive iteration"),
    ]  # type: List[Stripped]

    errors = []  # type: List[Error]

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            errors.append(
                Error(
                    cls.parsed.node,
                    f"NOTE (mristin, 2023-09-27): "
                    f"The class {cls.name!r} is marked as implementation specific. "
                    f"However, we currently do not generate the C++ iteration code "
                    f"over implementation-specific classes. Please contact "
                    f"the developers if you need this feature.",
                )
            )
            continue

        blocks.extend(_generate_iteration_over_cls(cls=cls))

    blocks.append(_generate_always_done_iterator())

    blocks.append(
        _generate_new_non_recursive_iterator_function(symbol_table=symbol_table)
    )

    blocks.append(Stripped("// endregion Non-recursive iteration"))

    blocks.append(Stripped("// region Recursive iteration"))

    blocks.extend(_generate_recursive_iteration())

    blocks.append(Stripped("// endregion Recursive iteration"))

    blocks.append(Stripped("// region Iterator facade"))

    blocks.extend(_generate_iterator_implementation())

    blocks.append(Stripped("// endregion Iterator facade"))

    blocks.append(Stripped("// region Descents"))

    blocks.extend(_generate_descent_and_descent_once_implementations())

    blocks.append(Stripped("// endregion Descents"))

    if len(symbol_table.enumerations) > 0:
        blocks.append(Stripped("// region Over enumerations"))

        for enum in symbol_table.enumerations:
            blocks.append(_generate_over_enum_implementation(enum))

        blocks.append(Stripped("// endregion Over enumerations"))

    if len(errors) > 0:
        return None, errors

    blocks.extend(
        [
            cpp_common.generate_namespace_closing(namespace),
            cpp_common.WARNING,
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


# endregion



"""Understand the hierarchy of classes in the symbol table as an ontology."""
from typing import (
    Sequence,
    Optional,
    Set,
    Final,
    Mapping,
    MutableMapping,
    List,
    cast,
    Tuple,
)

import sortedcontainers
from icontract import require, ensure

from aas_core_codegen import parse
from aas_core_codegen.common import Error, Identifier


def first_not_in_topological_order(
    classes: Sequence[parse.Class], parsed_symbol_table: parse.SymbolTable
) -> Optional[parse.Class]:
    """
    Verify that ``classes`` are topologically sorted.

    :return: The first class which is not fitting the expected order.
    """
    observed = set()  # type: Set[parse.Class]
    for cls in classes:
        for parent_name in cls.inheritances:
            # We ignore here the initial set of the constrained primitives.
            if parent_name in parse.PRIMITIVE_TYPES:
                assert len(cls.inheritances) == 1, (
                    f"A constrained primitive type in the initial set should only "
                    f"inherit from the primitive type. {cls.name=}"
                )
                continue

            parent = parsed_symbol_table.must_find_class(parent_name)
            if parent not in observed:
                return cls

        observed.add(cls)

    return None


class _UnverifiedOntology:
    """
    Provide an ontology computed from a symbol table.

    This private is explicitly made protected to signal that it has not been vetted
    yet and might be inconsistent. For example, there might be classes which have
    conflicting properties or methods with their ancestors.
    """

    #: Topologically sorted classes
    classes: Final[Sequence[parse.Class]]

    #: Map class 🠒 topologically sorted ancestors
    _ancestors_of: Final[Mapping[parse.Class, Sequence[parse.Class]]]

    _descendants_of: Final[Mapping[parse.Class, Sequence[parse.Class]]]

    # fmt: off
    @require(
        lambda classes, parsed_symbol_table:
        first_not_in_topological_order(classes, parsed_symbol_table) is None
    )
    @require(
        lambda classes:
        len(classes) == 0
        or sum(
            1
            for inheritance in classes[0].inheritances
            if inheritance not in parse.PRIMITIVE_TYPES
        ) == 0,
        "Origins first",
    )
    @require(
        lambda classes: len(set(classes)) == len(classes),
        "Unique classes in the topological sort",
    )
    @ensure(
        lambda self:
        all(
            self.can_list_descendants(cls)
            for cls in self.classes
        ),
        "The descendants` defined for all classes"
    )
    @ensure(
        lambda self:
        all(
            self.can_list_ancestors(cls)
            for cls in self.classes
        ),
        "The ancestors defined for all classes"
    )
    @ensure(
        lambda self, parsed_symbol_table: all(
            first_not_in_topological_order(
                class_ancestors, parsed_symbol_table) is None
            for class_ancestors in self._ancestors_of.values()
        )
    )
    # fmt: on
    def __init__(
        self, classes: Sequence[parse.Class], parsed_symbol_table: parse.SymbolTable
    ) -> None:
        """Initialize with the given values and pre-compute the ancestors."""
        self.classes = classes

        # region Determine ancestors

        ancestors_of = dict()  # type: MutableMapping[parse.Class, List[parse.Class]]

        order_of = {cls: i for i, cls in enumerate(classes)}

        for cls in classes:
            if any(
                parent_name in parse.PRIMITIVE_TYPES for parent_name in cls.inheritances
            ):
                assert len(cls.inheritances) == 1, (
                    f"A constrained primitive type in the initial set should only "
                    f"inherit from the primitive type. {cls.name=}"
                )
                ancestors_of[cls] = []

                continue

            parents = [
                parsed_symbol_table.must_find_class(parent_name)
                for parent_name in cls.inheritances
            ]

            parents_with_order = [(order_of[parent], parent) for parent in parents]

            sorted_parents = sorted(parents_with_order, key=lambda item: item[0])

            class_ancestors = []  # type: List[parse.Class]
            for _, parent in sorted_parents:
                assert parent in ancestors_of, (
                    f"Expected to process all the parent's of the class {cls.name} "
                    f"before (due to topological sort), "
                    f"but the parent class {parent.name} has not been processed"
                )

                class_ancestors.extend(ancestors_of[parent])

                class_ancestors.append(parent)

            ancestors_of[cls] = class_ancestors

        self._ancestors_of = ancestors_of

        # endregion

        # region Determine descendants

        descendants_of = dict()  # type: MutableMapping[parse.Class, List[parse.Class]]

        # Simply inverse

        for cls in self.classes:
            descendants_of[cls] = []

        for cls, ancestors in ancestors_of.items():
            for ancestor in ancestors:
                descendants_of[ancestor].append(cls)

        self._descendants_of = descendants_of

        # endregion

    def can_list_ancestors(self, cls: parse.Class) -> bool:
        """Return ``True`` if there is a record of the ``cls``'s ancestors."""
        return cls in self._ancestors_of

    def list_ancestors(self, cls: parse.Class) -> Sequence[parse.Class]:
        """Retrieve the ancestors of the given class ``cls``."""
        result = self._ancestors_of.get(cls, None)
        if result is None:
            raise KeyError(
                f"The ancestors of the class {cls} have not been precomputed."
            )

        return result

    def can_list_descendants(self, cls: parse.Class) -> bool:
        """Return ``True`` if there is a record of the ``cls``'s descendants."""
        return cls in self._descendants_of

    def list_descendants(self, cls: parse.Class) -> Sequence[parse.Class]:
        """Retrieve the descendants of the given class ``cls``."""

        result = self._descendants_of.get(cls, None)
        if result is None:
            raise KeyError(
                f"The descendants of the class {cls} have not been precomputed."
            )

        return result


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _topologically_sort(
    parsed_symbol_table: parse.SymbolTable,
) -> Tuple[Optional[_UnverifiedOntology], Optional[parse.Class]]:
    """
    Sort topologically all the classes in the ``parsed_symbol_table``.

    :return: topologically sorted classes, or a class in a cycle
    """
    # See https://en.wikipedia.org/wiki/Topological_sorting#Depth-first%20search
    # We use sorted containers to avoid non-deterministic behavior.

    result = []  # type: List[parse.Class]

    without_permanent_marks = sortedcontainers.SortedSet(
        key=lambda a_class: a_class.name
    )  # type: sortedcontainers.SortedSet[parse.Class]

    for our_type in parsed_symbol_table.our_types:
        if not isinstance(our_type, parse.Class):
            continue

        without_permanent_marks.add(our_type)

    permanent_marks = sortedcontainers.SortedSet(
        key=lambda a_class: a_class.name
    )  # type: sortedcontainers.SortedSet[parse.Class]

    temporary_marks = sortedcontainers.SortedSet(
        key=lambda a_class: a_class.name
    )  # type: sortedcontainers.SortedSet[parse.Class]

    visited_more_than_once = None  # type: Optional[parse.Class]

    def visit(cls: parse.Class) -> None:
        nonlocal visited_more_than_once
        nonlocal result

        if visited_more_than_once:
            return

        if cls in permanent_marks:
            return

        if cls in temporary_marks:
            visited_more_than_once = cls
            return

        temporary_marks.add(cls)

        for an_identifier in cls.inheritances:
            # We ignore the primitive types from which constrained primitive types
            # inherit.
            if an_identifier in parse.PRIMITIVE_TYPES:
                continue

            another_cls = parsed_symbol_table.must_find_class(an_identifier)

            visit(cls=another_cls)

        temporary_marks.remove(cls)
        permanent_marks.add(cls)

        if cls in without_permanent_marks:
            without_permanent_marks.remove(cls)

        result.append(cls)

    while len(without_permanent_marks) > 0 and not visited_more_than_once:
        visit(without_permanent_marks[0])

    if visited_more_than_once:
        return None, visited_more_than_once

    return (
        _UnverifiedOntology(classes=result, parsed_symbol_table=parsed_symbol_table),
        None,
    )


class Ontology(_UnverifiedOntology):
    """
    Provide an ontology computed from a symbol table.

    The ontology has been verified through :py:func:`map_symbol_table_to_ontology`.
    """


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def map_symbol_table_to_ontology(
    parsed_symbol_table: parse.SymbolTable,
) -> Tuple[Optional[Ontology], Optional[List[Error]]]:
    """Infer the ontology of the classes from the ``parsed_symbol_table``."""
    ontology, visited_more_than_once = _topologically_sort(
        parsed_symbol_table=parsed_symbol_table
    )
    if visited_more_than_once is not None:
        return (
            None,
            [
                Error(
                    visited_more_than_once.node,
                    f"Expected no cycles in the inheritance, "
                    f"but the class {visited_more_than_once.name} has been observed "
                    f"in a cycle",
                )
            ],
        )

    assert ontology is not None

    errors = []  # type: List[Error]

    # region Check that properties and methods do not conflict among ancestors

    for our_type in parsed_symbol_table.our_types:
        if not isinstance(our_type, parse.Class):
            continue

        ancestors = ontology.list_ancestors(cls=our_type)

        observed_properties = dict()  # type: MutableMapping[str, parse.Class]
        observed_methods = dict()  # type: MutableMapping[str, parse.Class]

        for ancestor in ancestors:
            for prop in ancestor.properties:
                if prop.name not in observed_properties:
                    observed_properties[prop.name] = ancestor

            for method in ancestor.methods:
                if method.name not in observed_properties:
                    observed_methods[method.name] = ancestor

        for prop in our_type.properties:
            if prop.name in observed_properties:
                errors.append(
                    Error(
                        prop.node,
                        f"The property has already been defined in the ancestor "
                        f"class {observed_properties[prop.name].name}: {prop.name}",
                    )
                )

        for method in our_type.methods:
            if method.name == "__init__":
                continue

            if method.name in observed_methods:
                errors.append(
                    Error(
                        method.node,
                        f"The method has already been defined in the ancestor "
                        f"class {observed_methods[method.name].name}: {method.name}",
                    )
                )

    # endregion

    # region Check that ancestors do not have constructors if the class lacks one

    for our_type in parsed_symbol_table.our_types:
        if not isinstance(our_type, parse.Class):
            continue

        if "__init__" not in our_type.methods_by_name:
            for ancestor in ontology.list_ancestors(our_type):
                ancestor_init = ancestor.methods_by_name.get(
                    Identifier("__init__"), None
                )

                if ancestor_init is not None and len(ancestor_init.arguments) > 1:
                    argument_names_str = ", ".join(
                        arg.name for arg in ancestor_init.arguments
                    )

                    errors.append(
                        Error(
                            our_type.node,
                            f"The class {our_type.name} does not specify "
                            f"a constructor, but the ancestor class "
                            f"{ancestor.name} specifies a constructor with "
                            f"arguments: {argument_names_str}",
                        )
                    )

    # endregion

    if len(errors) > 0:
        return None, errors

    return cast(Ontology, ontology), None



"""Generate C# code for JSON-ization based on the intermediate representation."""

import io
import textwrap
from typing import Tuple, Optional, List

from icontract import ensure

from aas_core_codegen import intermediate, naming, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    Identifier,
    assert_never,
    indent_but_first_line,
)
from aas_core_codegen.csharp import (
    common as csharp_common,
    naming as csharp_naming,
)
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


def _generate_from_method_for_enumeration(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    """Generate the deserialization method for an enumeration."""
    name = csharp_naming.enum_name(identifier=enumeration.name)

    message_literal = csharp_common.string_literal(
        f"Not a valid JSON representation of {name}"
    )

    return Stripped(
        f"""\
/// <summary>
/// Deserialize the enumeration {name} from the <paramref name="node" />.
/// </summary>
/// <param name="node">JSON node to be parsed</param>
/// <param name="error">Error, if any, during the deserialization</param>
internal static Aas.{name}? {name}From(
{I}Nodes.JsonNode node,
{I}out Reporting.Error? error)
{{
{I}error = null;
{I}string? text = DeserializeImplementation.StringFrom(
{II}node, out error);
{I}if (error != null)
{I}{{
{II}return null;
{I}}}
{I}if (text == null)
{I}{{
{II}throw new System.InvalidOperationException(
{III}"Unexpected text null if error null");
{I}}}
{I}Aas.{name}? result = Stringification.{name}FromString(text);
{I}if (result == null)
{I}{{
{II}error = new Reporting.Error(
{III}{message_literal});
{I}}}
{I}return result;
}}  // internal static {name}From"""
    )


def _generate_from_method_for_interface(
    interface: intermediate.Interface,
) -> Stripped:
    """Generate the deserialization method for an interface."""
    name = csharp_naming.interface_name(interface.name)

    blocks = [
        Stripped("error = null;"),
        Stripped(
            f"""\
var obj = node as Nodes.JsonObject;
if (obj == null)
{{
{I}error = new Reporting.Error(
{II}$"Expected Nodes.JsonObject, but got {{node.GetType()}}");
{I}return null;
}}"""
        ),
        Stripped(
            f"""\
Nodes.JsonNode? modelTypeNode = obj["modelType"];
if (modelTypeNode == null)
{{
{I}error = new Reporting.Error(
{II}"Expected a model type, but none is present");
{I}return null;
}}
Nodes.JsonValue? modelTypeValue = modelTypeNode as Nodes.JsonValue;
if (modelTypeValue == null)
{{
{I}error = new Reporting.Error(
{II}"Expected JsonValue, " +
{II}$"but got {{modelTypeNode.GetType()}}");
{I}return null;
}}
modelTypeValue.TryGetValue<string>(out string? modelType);
if (modelType == null)
{{
{I}error = new Reporting.Error(
{II}"Expected a string, " +
{II}$"but the conversion failed from {{modelTypeValue}}");
{I}return null;
}}"""
        ),
    ]  # type: List[Stripped]

    # region Write the switch block

    switch_writer = io.StringIO()
    switch_writer.write(
        """\
switch (modelType)
{
"""
    )

    for implementer in interface.implementers:
        model_type = naming.json_model_type(implementer.name)
        implementer_name = csharp_naming.class_name(implementer.name)
        switch_writer.write(
            f"""\
{I}case {csharp_common.string_literal(model_type)}:
{II}return {implementer_name}From(
{III}node, out error);
"""
        )

    switch_writer.write(
        f"""\
{I}default:
{II}error = new Reporting.Error(
{III}$"Unexpected model type for {name}: {{modelType}}");
{II}return null;
}}"""
    )
    blocks.append(Stripped(switch_writer.getvalue()))

    # endregion

    writer = io.StringIO()

    writer.write(
        f"""\
/// <summary>
/// Deserialize an instance of {name} by dispatching
/// based on <c>modelType</c> property of the <paramref name="node" />.
/// </summary>
/// <param name="node">JSON node to be parsed</param>
/// <param name="error">Error, if any, during the deserialization</param>
[CodeAnalysis.SuppressMessage("ReSharper", "InconsistentNaming")]
public static Aas.{name}? {name}From(
{I}Nodes.JsonNode node,
{I}out Reporting.Error? error)
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write(f"\n}}  // public static Aas.{name} {name}From")

    return Stripped(writer.getvalue())


_PARSE_METHOD_BY_PRIMITIVE_TYPE = {
    intermediate.PrimitiveType.BOOL: "DeserializeImplementation.BoolFrom",
    intermediate.PrimitiveType.INT: "DeserializeImplementation.LongFrom",
    intermediate.PrimitiveType.FLOAT: "DeserializeImplementation.DoubleFrom",
    intermediate.PrimitiveType.STR: "DeserializeImplementation.StringFrom",
    intermediate.PrimitiveType.BYTEARRAY: "DeserializeImplementation.BytesFrom",
}
assert all(
    literal in _PARSE_METHOD_BY_PRIMITIVE_TYPE for literal in intermediate.PrimitiveType
)


def _parse_method_for_atomic_value(
    type_annotation: intermediate.AtomicTypeAnnotation,
) -> Stripped:
    """Determine the parse method for deserializing an atomic non-optional value."""
    parse_method: str

    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        parse_method = _PARSE_METHOD_BY_PRIMITIVE_TYPE[type_annotation.a_type]

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type
        if isinstance(our_type, intermediate.Enumeration):
            enum_name = csharp_naming.enum_name(our_type.name)
            parse_method = f"DeserializeImplementation.{enum_name}From"

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            parse_method = _PARSE_METHOD_BY_PRIMITIVE_TYPE[our_type.constrainee]

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if our_type.interface is not None:
                interface_name = csharp_naming.interface_name(our_type.interface.name)
                parse_method = f"DeserializeImplementation.{interface_name}From"
            else:
                cls_name = csharp_naming.class_name(our_type.name)
                parse_method = f"DeserializeImplementation.{cls_name}From"

        else:
            assert_never(our_type)
    else:
        assert_never(type_annotation)

    return Stripped(parse_method)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_deserialize_constructor_argument(
    arg: intermediate.Argument,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the code snippet for de-serializing the constructor argument ``arg``."""
    type_anno = intermediate.beneath_optional(arg.type_annotation)

    # Prefix the variables to avoid naming conflicts
    target_var = csharp_naming.variable_name(Identifier(f"the_{arg.name}"))

    json_name = naming.json_property(arg.name)
    assert not csharp_common.needs_escaping(json_name)

    json_literal = csharp_common.string_literal(json_name)

    parse_block: Stripped

    if isinstance(
        type_anno,
        (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
    ):
        parse_method = _parse_method_for_atomic_value(type_anno)

        parse_block = Stripped(
            f"""\
{target_var} = {parse_method}(
{I}keyValue.Value,
{I}out error);
if (error != null)
{{
{I}error.PrependSegment(
{II}new Reporting.NameSegment(
{III}{json_literal}));
{I}return null;
}}
if ({target_var} == null)
{{
{I}throw new System.InvalidOperationException(
{II}"Unexpected {target_var} null when error is also null");
}}"""
        )

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        assert not isinstance(
            type_anno.items,
            (intermediate.OptionalTypeAnnotation, intermediate.ListTypeAnnotation),
        ), (
            "We chose to implement only a very limited pattern matching; "
            "see intermediate._translate_._verify_only_simple_type_patterns"
        )

        item_type = csharp_common.generate_type(type_anno.items)

        array_var = csharp_naming.variable_name(Identifier(f"array_{arg.name}"))
        index_var = csharp_naming.variable_name(Identifier(f"index_{arg.name}"))

        parse_method = _parse_method_for_atomic_value(type_anno.items)

        parse_block = Stripped(
            f"""\
Nodes.JsonArray? {array_var} = keyValue.Value as Nodes.JsonArray;
if ({array_var} == null)
{{
{I}error = new Reporting.Error(
{II}$"Expected a JsonArray, but got {{keyValue.Value.GetType()}}");
{I}error.PrependSegment(
{II}new Reporting.NameSegment(
{III}{json_literal}));
{I}return null;
}}
{target_var} = new List<{item_type}>(
{I}{array_var}.Count);
int {index_var} = 0;
foreach (Nodes.JsonNode? item in {array_var})
{{
{I}if (item == null)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected a non-null item, but got a null");
{II}error.PrependSegment(
{III}new Reporting.IndexSegment(
{IIII}{index_var}));
{II}error.PrependSegment(
{III}new Reporting.NameSegment(
{IIII}{json_literal}));
{II}return null;
{I}}}
{I}{item_type}? parsedItem = {parse_method}(
{II}item ?? throw new System.InvalidOperationException(),
{II}out error);
{I}if (error != null)
{I}{{
{II}error.PrependSegment(
{III}new Reporting.IndexSegment(
{IIII}{index_var}));
{II}error.PrependSegment(
{III}new Reporting.NameSegment(
{IIII}{json_literal}));
{II}return null;
{I}}}
{I}{target_var}.Add(
{II}parsedItem
{III}?? throw new System.InvalidOperationException(
{IIII}"Unexpected result null when error is null"));
{I}{index_var}++;
}}"""
        )
    else:
        assert_never(arg.type_annotation)

    # NOTE (mristin):
    # We need to add a prologue to the parsing body to explicitly check for null
    # values as the null values are not allowed for optional properties by
    # specification.
    if isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation):
        parse_block = Stripped(
            f"""\
if (keyValue.Value == null)
{{
{I}error = new Reporting.Error(
{II}"Expected optional property to be absent, " +
{II}"but got null instead");
{I}error.PrependSegment(
{II}new Reporting.NameSegment(
{III}{json_literal}));
{I}return null;
}}

{parse_block}"""
        )
    else:
        parse_block = Stripped(
            f"""\
if (keyValue.Value == null)
{{
{I}error = new Reporting.Error(
{II}"Unexpected null for a required property");
{I}error.PrependSegment(
{II}new Reporting.NameSegment(
{III}{json_literal}));
{I}return null;
}}

{parse_block}"""
        )

    return parse_block, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_from_method_for_class(
    cls: intermediate.ConcreteClass,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the deserialization method for a concrete class."""
    errors = []  # type: List[Error]

    name = csharp_naming.class_name(cls.name)

    blocks = [
        Stripped("error = null;"),
        Stripped(
            f"""\
Nodes.JsonObject? obj = node as Nodes.JsonObject;
if (obj == null)
{{
{I}error = new Reporting.Error(
{II}$"Expected a JsonObject, but got {{node.GetType()}}");
{I}return null;
}}"""
        ),
    ]  # type: List[Stripped]

    # region Initialize argument variables to null

    args_init_writer = io.StringIO()
    for i, arg in enumerate(cls.constructor.arguments):
        arg_var = csharp_naming.variable_name(Identifier(f"the_{arg.name}"))
        arg_type = csharp_common.generate_type(arg.type_annotation)

        # NOTE (mristin, 2022-07-22):
        # We make all the argument variables optional since we switch over
        # the properties. Even the mandatory constructor arguments can be omitted
        # during an invalid deserialization!
        if not arg_type.endswith("?"):
            arg_type = Stripped(f"{arg_type}?")

        if i > 0:
            args_init_writer.write("\n")
        args_init_writer.write(f"{arg_type} {arg_var} = null;")

    blocks.append(Stripped(args_init_writer.getvalue()))

    if cls.serialization.with_model_type:
        blocks.append(Stripped("string? modelType = null;"))

    # endregion

    # region Switch on property name

    cases = []  # type: List[Stripped]
    for arg in cls.constructor.arguments:
        case_body, error = _generate_deserialize_constructor_argument(arg=arg)
        if error is not None:
            errors.append(error)
        else:
            assert case_body is not None
            json_name = naming.json_property(arg.name)

            cases.append(
                Stripped(
                    f"""\
case {csharp_common.string_literal(json_name)}:
{{
{I}{indent_but_first_line(case_body, I)}
{I}break;
}}"""
                )
            )

    if len(errors) > 0:
        return None, errors

    if cls.serialization.with_model_type:
        model_type = naming.json_model_type(cls.name)

        cases.append(
            Stripped(
                f"""\
case "modelType":
{I}{{
{II}if (keyValue.Value == null)
{II}{{
{III}error = new Reporting.Error(
{IIII}"Expected a model type, but got null");
{III}return null;
{II}}}
{II}modelType = DeserializeImplementation.StringFrom(
{III}keyValue.Value,
{III}out error);
{II}if (error != null)
{II}{{
{III}error.PrependSegment(
{IIII}new Reporting.NameSegment(
{IIIII}"modelType"));
{III}return null;
{II}}}

{II}if (modelType != "{model_type}")
{II}{{
{III}error = new Reporting.Error(
{IIII}"Expected the model type '{model_type}', " +
{IIII}$"but got {{modelType}}");
{III}error.PrependSegment(
{IIII}new Reporting.NameSegment(
{IIIII}"modelType"));
{III}return null;
{II}}}
{II}break;
{I}}}"""
            )
        )

    cases.append(
        Stripped(
            f"""\
default:
{I}error = new Reporting.Error(
{II}$"Unexpected property: {{keyValue.Key}}");
{I}return null;"""
        )
    )

    foreach_writer = io.StringIO()
    foreach_writer.write(
        f"""\
foreach (var keyValue in obj)
{{
{I}switch (keyValue.Key)
{I}{{"""
    )

    for case_block in cases:
        foreach_writer.write("\n")
        foreach_writer.write(textwrap.indent(case_block, II))

    foreach_writer.write(f"\n{I}}}\n}}")

    blocks.append(Stripped(foreach_writer.getvalue()))

    # endregion

    # region Check required

    required_check_writer = io.StringIO()
    for i, arg in enumerate(cls.constructor.arguments):
        if isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        arg_var = csharp_naming.variable_name(Identifier(f"the_{arg.name}"))
        json_name = naming.json_property(arg.name)
        assert not csharp_common.needs_escaping(json_name)

        if i > 0:
            required_check_writer.write("\n\n")

        required_check_writer.write(
            f"""\
if ({arg_var} == null)
{{
{I}error = new Reporting.Error(
{II}"Required property \\"{json_name}\\" is missing");
{I}return null;
}}"""
        )

    blocks.append(Stripped(required_check_writer.getvalue()))

    if cls.serialization.with_model_type:
        blocks.append(
            Stripped(
                f"""\
if (modelType == null)
{{
{I}error = new Reporting.Error(
{II}"Required property \\"modelType\\" is missing");
{I}return null;
}}"""
            )
        )

    # endregion

    # region Pass in arguments to the constructor

    property_names = [prop.name for prop in cls.properties]
    constructor_argument_names = [arg.name for arg in cls.constructor.arguments]

    # fmt: off
    assert (
            set(prop.name for prop in cls.properties)
            == set(arg.name for arg in cls.constructor.arguments)
    ), (
        f"Expected the properties to coincide with constructor arguments, "
        f"but they do not for {cls.name!r}:"
        f"{property_names=}, {constructor_argument_names=}"
    )
    # fmt: on

    if len(cls.constructor.arguments) == 0:
        blocks.append(Stripped(f"return new Aas.{name}();"))
    else:
        init_writer = io.StringIO()
        init_writer.write(f"return new Aas.{name}(\n")

        for i, arg in enumerate(cls.constructor.arguments):
            prop = cls.properties_by_name[arg.name]

            # NOTE (mristin, 2022-03-11):
            # The argument to the constructor may be optional while the property
            # might be required, since we can set the default value in the body of
            # the constructor. However, we can not have an optional property and a
            # required constructor argument as we then would not know how to create
            # the instance.

            if not (
                intermediate.type_annotations_equal(
                    arg.type_annotation, prop.type_annotation
                )
                or intermediate.type_annotations_equal(
                    intermediate.beneath_optional(arg.type_annotation),
                    prop.type_annotation,
                )
            ):
                errors.append(
                    Error(
                        arg.parsed.node,
                        f"Expected type annotation for property {prop.name!r} "
                        f"and constructor argument {arg.name!r} "
                        f"of the class {cls.name!r} to have matching types, "
                        f"but they do not: "
                        f"property type is {prop.type_annotation} "
                        f"and argument type is {arg.type_annotation}. "
                        f"Hence we do not know how to generate the call "
                        f"to the constructor in the JSON de-serialization.",
                    )
                )
                continue

            arg_var = csharp_naming.variable_name(Identifier(f"the_{arg.name}"))

            init_writer.write(f"{I}{arg_var}")
            if not isinstance(
                prop.type_annotation, intermediate.OptionalTypeAnnotation
            ):
                init_writer.write("\n")

                init_writer.write(
                    f"""\
{II} ?? throw new System.InvalidOperationException(
{III}"Unexpected null, had to be handled before")"""
                )

            if i < len(cls.constructor.arguments) - 1:
                init_writer.write(",\n")
            else:
                init_writer.write(");")

        if len(errors) > 0:
            return None, errors

        blocks.append(Stripped(init_writer.getvalue()))
    # endregion

    writer = io.StringIO()

    writer.write(
        f"""\
/// <summary>
/// Deserialize an instance of {name} from <paramref name="node" />.
/// </summary>
/// <param name="node">JSON node to be parsed</param>
/// <param name="error">Error, if any, during the deserialization</param>
internal static Aas.{name}? {name}From(
{I}Nodes.JsonNode node,
{I}out Reporting.Error? error)
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write(f"\n}}  // internal static {name}From")

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_deserialize_impl(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the implementation of the deserialization."""
    errors = []  # type: List[Error]

    blocks = [
        Stripped(
            f"""\
/// <summary>Convert <paramref name="node" /> to a boolean.</summary>
/// <param name="node">JSON node to be parsed</param>
/// <param name="error">Error, if any, during the deserialization</param>
internal static bool? BoolFrom(
{I}Nodes.JsonNode node,
{I}out Reporting.Error? error)
{{
{I}error = null;
{I}Nodes.JsonValue? value = node as Nodes.JsonValue;
{I}if (value == null)
{I}{{
{II}error = new Reporting.Error(
{III}$"Expected a JsonValue, but got {{node.GetType()}}");
{II}return null;
{I}}}
{I}bool ok = value.TryGetValue<bool>(out bool result);
{I}if (!ok)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected a boolean, but the conversion failed " +
{III}$"from {{value.ToJsonString()}}");
{II}return null;
{I}}}
{I}return result;
}}"""
        ),
        Stripped(
            f"""\
/// <summary>
/// Convert the <paramref name="node" /> to a long 64-bit integer.
/// </summary>
/// <param name="node">JSON node to be parsed</param>
/// <param name="error">Error, if any, during the deserialization</param>
internal static long? LongFrom(
{I}Nodes.JsonNode node,
{I}out Reporting.Error? error)
{{
{I}error = null;
{I}Nodes.JsonValue? value = node as Nodes.JsonValue;
{I}if (value == null)
{I}{{
{II}error = new Reporting.Error(
{III}$"Expected a JsonValue, but got {{node.GetType()}}");
{II}return null;
{I}}}
{I}bool ok = value.TryGetValue<long>(out long result);
{I}if (!ok)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected a 64-bit long integer, but the conversion failed " +
{III}$"from {{value.ToJsonString()}}");
{II}return null;
{I}}}
{I}return result;
}}"""
        ),
        Stripped(
            f"""\
/// <summary>
/// Convert the <paramref name="node" /> to a double-precision 64-bit float.
/// </summary>
/// <param name="node">JSON node to be parsed</param>
/// <param name="error">Error, if any, during the deserialization</param>
internal static double? DoubleFrom(
{I}Nodes.JsonNode node,
{I}out Reporting.Error? error)
{{
{I}error = null;
{I}Nodes.JsonValue? value = node as Nodes.JsonValue;
{I}if (value == null)
{I}{{
{II}error = new Reporting.Error(
{III}$"Expected a JsonValue, but got {{node.GetType()}}");
{II}return null;
{I}}}
{I}bool ok = value.TryGetValue<double>(out double result);
{I}if (!ok)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected a 64-bit double-precision float, " +
{III}"but the conversion failed " +
{III}$"from {{value.ToJsonString()}}");
{II}return null;
{I}}}
{I}return result;
}}"""
        ),
        Stripped(
            f"""\
/// <summary>
/// Convert the <paramref name="node" /> to a string.
/// </summary>
/// <param name="node">JSON node to be parsed</param>
/// <param name="error">Error, if any, during the deserialization</param>
internal static string? StringFrom(
{I}Nodes.JsonNode node,
{I}out Reporting.Error? error)
{{
{I}error = null;
{I}Nodes.JsonValue? value = node as Nodes.JsonValue;
{I}if (value == null)
{I}{{
{II}error = new Reporting.Error(
{III}$"Expected a JsonValue, but got {{node.GetType()}}");
{II}return null;
{I}}}
{I}bool ok = value.TryGetValue<string>(out string? result);
{I}if (!ok)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected a string, but the conversion failed " +
{III}$"from {{value.ToJsonString()}}");
{II}return null;
{I}}}
{I}if (result == null)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected a string, but got a null");
{II}return null;
{I}}}
{I}return result;
}}"""
        ),
        Stripped(
            f"""\
/// <summary>
/// Convert the <paramref name="node" /> to bytes.
/// </summary>
/// <param name="node">JSON node to be parsed</param>
/// <param name="error">Error, if any, during the deserialization</param>
internal static byte[]? BytesFrom(
{I}Nodes.JsonNode node,
{I}out Reporting.Error? error)
{{
{I}error = null;
{I}Nodes.JsonValue? value = node as Nodes.JsonValue;
{I}if (value == null)
{I}{{
{II}error = new Reporting.Error(
{III}$"Expected a JsonValue, but got {{node.GetType()}}");
{II}return null;
{I}}}
{I}bool ok = value.TryGetValue<string>(out string? text);
{I}if (!ok)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected a string, but the conversion failed " +
{III}$"from {{value.ToJsonString()}}");
{II}return null;
{I}}}
{I}if (text == null)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected a string, but got a null");
{II}return null;
{I}}}
{I}try
{I}{{
{II}return System.Convert.FromBase64String(text);
{I}}}
{I}catch (System.FormatException exception)
{I}{{
{II}error = new Reporting.Error(
{III}"Expected Base-64 encoded bytes, but the conversion failed " +
{III}$"because: {{exception}}");
{II}return null;
{I}}}
}}"""
        ),
    ]  # type: List[Stripped]

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            blocks.append(_generate_from_method_for_enumeration(enumeration=our_type))

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            continue

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if our_type.interface is not None:
                blocks.append(
                    _generate_from_method_for_interface(interface=our_type.interface)
                )

            if isinstance(our_type, intermediate.ConcreteClass):
                if our_type.is_implementation_specific:
                    implementation_key = specific_implementations.ImplementationKey(
                        f"Jsonization/DeserializeImplementation/{our_type.name}_from.cs"
                    )

                    implementation = spec_impls.get(implementation_key, None)
                    if implementation is None:
                        errors.append(
                            Error(
                                our_type.parsed.node,
                                f"The jsonization snippet is missing "
                                f"for the implementation-specific "
                                f"class {our_type.name}: {implementation_key}",
                            )
                        )
                        continue

                    blocks.append(spec_impls[implementation_key])
                else:
                    block, cls_errors = _generate_from_method_for_class(cls=our_type)
                    if cls_errors is not None:
                        errors.extend(cls_errors)
                        continue
                    else:
                        assert block is not None
                        blocks.append(block)
        else:
            assert_never(our_type)

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()

    writer.write(
        """\
/// <summary>
/// Implement the deserialization of meta-model classes from JSON nodes.
/// </summary>
/// <remarks>
/// The implementation propagates an <see cref="Reporting.Error" /> instead of relying
/// on exceptions. Under the assumption that incorrect data is much less
/// frequent than correct data, this makes the deserialization more
/// efficient.
///
/// However, we do not want to force the client to deal with
/// the <see cref="Reporting.Error" /> class as this is not intuitive. Therefore
/// we distinguish the implementation, realized in
/// <see cref="DeserializeImplementation" />, and the facade given in
/// <see cref="Deserialize" /> class.
/// </remarks>
internal static class DeserializeImplementation
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public static class DeserializeImplementation")

    return Stripped(writer.getvalue()), None


def _generate_deserialize_from(name: str) -> Stripped:
    """Generate the facade deserialization method for the type with C# ``name``."""
    writer = io.StringIO()
    writer.write(
        f"""\
/// <summary>
/// Deserialize an instance of {name} from <paramref name="node" />.
/// </summary>
/// <param name="node">JSON node to be parsed</param>
/// <exception cref="Jsonization.Exception">
/// Thrown when <paramref name="node" /> is not a valid JSON
/// representation of {name}.
/// </exception>
"""
    )

    if name.startswith("I"):
        writer.write(
            '[CodeAnalysis.SuppressMessage("ReSharper", "InconsistentNaming")]\n'
        )

    writer.write(
        f"""\
public static Aas.{name} {name}From(
{I}Nodes.JsonNode node)
{{
{I}Aas.{name}? result = DeserializeImplementation.{name}From(
{II}node,
{II}out Reporting.Error? error);
{I}if (error != null)
{I}{{
{II}throw new Jsonization.Exception(
{III}Reporting.GenerateJsonPath(error.PathSegments),
{III}error.Cause);
{I}}}
{I}return result
{II}?? throw new System.InvalidOperationException(
{III}"Unexpected output null when error is null");
}}"""
    )

    return Stripped(writer.getvalue())


def _generate_deserialize(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the deserializer with a deserialization method for each class."""
    blocks = []  # type: List[Stripped]
    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            blocks.append(
                _generate_deserialize_from(name=csharp_naming.enum_name(our_type.name))
            )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            continue

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if our_type.interface is not None:
                blocks.append(
                    _generate_deserialize_from(
                        name=csharp_naming.interface_name(our_type.interface.name)
                    )
                )

            if isinstance(our_type, intermediate.ConcreteClass):
                blocks.append(
                    _generate_deserialize_from(
                        name=csharp_naming.class_name(our_type.name)
                    )
                )
        else:
            assert_never(our_type)

    writer = io.StringIO()

    writer.write(
        """\
/// <summary>
/// Deserialize instances of meta-model classes from JSON nodes.
/// </summary>
"""
    )

    first_cls = (
        symbol_table.classes[0] if len(symbol_table.classes) > 0 else None
    )  # type: Optional[intermediate.ClassUnion]

    if first_cls is not None:
        cls_name: str

        if isinstance(first_cls, intermediate.AbstractClass):
            cls_name = csharp_naming.interface_name(first_cls.name)
        elif isinstance(first_cls, intermediate.ConcreteClass):
            cls_name = csharp_naming.class_name(first_cls.name)
        else:
            assert_never(first_cls)

        an_instance_variable = csharp_naming.variable_name(Identifier("an_instance"))

        writer.write(
            f"""\
/// <example>
/// Here is an example how to parse an instance of {cls_name}:
/// <code>
/// string someString = "... some JSON ...";
/// var node = System.Text.Json.Nodes.JsonNode.Parse(someString);
/// Aas.{cls_name} {an_instance_variable} = Deserialize.{cls_name}From(
/// {I}node);
/// </code>
/// </example>
"""
        )

    writer.write(
        """\
public static class Deserialize
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public static class Deserialize")

    return Stripped(writer.getvalue())


def _generate_serialize_primitive_value(
    primitive_type: intermediate.PrimitiveType, source_expr: Stripped
) -> Stripped:
    """
    Generate the snippet to serialize ``source_expr`` to JSON.

    Source expression is expected to be of ``primitive_type``.
    """
    if (
        primitive_type is intermediate.PrimitiveType.BOOL
        or primitive_type is intermediate.PrimitiveType.FLOAT
        or primitive_type is intermediate.PrimitiveType.STR
    ):
        # We can not use textwrap due to indent_but_first_line.
        return Stripped(
            f"""\
Nodes.JsonValue.Create(
{I}{indent_but_first_line(source_expr, I)})"""
        )
    elif primitive_type is intermediate.PrimitiveType.INT:
        # We can not use textwrap due to indent_but_first_line.
        return Stripped(
            f"""\
Transformer.ToJsonValue(
{I}{indent_but_first_line(source_expr, I)})"""
        )
    elif primitive_type is intermediate.PrimitiveType.BYTEARRAY:
        # We can not use textwrap due to indent_but_first_line.
        return Stripped(
            f"""\
Nodes.JsonValue.Create(
{I}System.Convert.ToBase64String(
{II}{indent_but_first_line(source_expr, II)}))"""
        )
    else:
        assert_never(primitive_type)


def _generate_serialize_atomic_value(
    type_annotation: intermediate.AtomicTypeAnnotation, source_expr: Stripped
) -> Stripped:
    """Generate the snippet to serialize ``source_expr`` to JSON."""
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return _generate_serialize_primitive_value(
            primitive_type=type_annotation.a_type, source_expr=source_expr
        )
    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type
        if isinstance(our_type, intermediate.Enumeration):
            name = csharp_naming.enum_name(our_type.name)

            # We can not use textwrap due to indent_but_first_line.
            return Stripped(
                f"""\
Serialize.{name}ToJsonValue(
{I}{indent_but_first_line(source_expr, I)})"""
            )
        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return _generate_serialize_primitive_value(
                primitive_type=our_type.constrainee, source_expr=source_expr
            )
        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            # We can not use textwrap due to indent_but_first_line.
            return Stripped(
                f"""\
Transform(
{I}{indent_but_first_line(source_expr, I)})"""
            )
        else:
            assert_never(our_type)
    else:
        assert_never(type_annotation)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_transform_property(
    prop: intermediate.Property,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the snippet to transform a property into a JSON node."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    stmts = []  # type: List[Stripped]

    name = csharp_naming.property_name(prop.name)
    prop_literal = csharp_common.string_literal(naming.json_property(prop.name))

    # NOTE (mristin, 2022-03-12):
    # For some unexplainable reason, C# compiler can not infer that properties which
    # are enumerations are not null after an ``if (that.someProperty != null)``.
    # Hence, we need to add a null-coalescing for these particular cases.
    # Otherwise, we can just stick to ``that.someProperty``.

    needs_null_coalescing = (
        isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation)
        and isinstance(prop.type_annotation.value, intermediate.OurTypeAnnotation)
        and isinstance(prop.type_annotation.value.our_type, intermediate.Enumeration)
    )
    if needs_null_coalescing:
        source_expr = Stripped("value")
    else:
        source_expr = Stripped(f"that.{name}")

    if isinstance(
        type_anno,
        (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
    ):
        conversion_expr = _generate_serialize_atomic_value(
            type_annotation=type_anno, source_expr=source_expr
        )
        stmts.append(Stripped(f"result[{prop_literal}] = {conversion_expr};"))
    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        assert not isinstance(
            type_anno.items,
            (intermediate.OptionalTypeAnnotation, intermediate.ListTypeAnnotation),
        ), (
            "We chose to implement only a very limited pattern matching; "
            "see intermediate._translate._verify_only_simple_type_patterns."
        )

        item_type = csharp_common.generate_type(type_anno.items)
        array_var = csharp_naming.variable_name(Identifier(f"array_{prop.name}"))

        item_conversion_expr = _generate_serialize_atomic_value(
            type_annotation=type_anno.items, source_expr=Stripped("item")
        )

        # We can not use textwrap due to indent_but_first_line.
        stmts.append(
            Stripped(
                f"""\
var {array_var} = new Nodes.JsonArray();
foreach ({item_type} item in {source_expr})
{{
{I}{array_var}.Add(
{II}{indent_but_first_line(item_conversion_expr, II)});
}}
result[{prop_literal}] = {array_var};"""
            )
        )
    else:
        assert_never(type_anno)

    serialize_block = Stripped("\n".join(stmts))
    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        if needs_null_coalescing:
            value_type = csharp_common.generate_type(prop.type_annotation.value)
            if isinstance(prop.type_annotation.value, intermediate.OurTypeAnnotation):
                our_type = prop.type_annotation.value.our_type
                if isinstance(
                    our_type,
                    (
                        intermediate.Enumeration,
                        intermediate.AbstractClass,
                        intermediate.ConcreteClass,
                    ),
                ):
                    value_type = Stripped(f"Aas.{value_type}")

            return (
                Stripped(
                    f"""\
if (that.{name} != null)
{{
{I}// We need to help the static analyzer with a null coalescing.
{I}{value_type} value = that.{name}
{II}?? throw new System.InvalidOperationException();
{I}{indent_but_first_line(serialize_block, I)}
}}"""
                ),
                None,
            )

        else:
            return (
                Stripped(
                    f"""\
if (that.{name} != null)
{{
{I}{indent_but_first_line(serialize_block, I)}
}}"""
                ),
                None,
            )
    else:
        return serialize_block, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_transform_for_class(
    cls: intermediate.ConcreteClass,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the transform method to a JSON object for the given concrete class."""
    errors = []  # type: List[Error]

    blocks = [Stripped("var result = new Nodes.JsonObject();")]  # type: List[Stripped]

    for prop in cls.properties:
        block, error = _generate_transform_property(prop=prop)
        if error is not None:
            errors.append(error)
        else:
            assert block is not None
            blocks.append(block)

    if len(errors) > 0:
        return None, errors

    if cls.serialization is not None and cls.serialization.with_model_type:
        model_type = csharp_common.string_literal(naming.json_model_type(cls.name))
        blocks.append(Stripped(f'result["modelType"] = {model_type};'))

    blocks.append(Stripped("return result;"))

    writer = io.StringIO()

    interface_name = csharp_naming.interface_name(cls.name)
    transform_name = csharp_naming.method_name(Identifier(f"transform_{cls.name}"))

    writer.write(
        f"""\
public override Nodes.JsonObject {transform_name}(
{I}Aas.{interface_name} that
)
{{
"""
    )

    for i, stmt in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(stmt, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_transformer(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate a transformer which transforms instances of the meta-model to JSON."""
    errors = []  # type: List[Error]

    blocks = [
        Stripped(
            f"""\
/// <summary>
/// Convert <paramref name="that" /> 64-bit long integer to a JSON value.
/// </summary>
/// <param name="that">value to be converted</param>
/// <exception name="System.ArgumentException">
/// Thrown if <paramref name="that" /> is not within the range where it
/// can be losslessly converted to a double floating number.
/// </exception>
[CodeAnalysis.SuppressMessage("ReSharper", "UnusedMember.Local")]
private static Nodes.JsonValue ToJsonValue(long that)
{{
{I}// We need to check that we can perform a lossless conversion.
{I}if ((long)((double)that) != that)
{I}{{
{II}throw new System.ArgumentException(
{III}$"The number can not be losslessly represented in JSON: {{that}}");
{I}}}
{I}return Nodes.JsonValue.Create(that);
}}"""
        ),
    ]  # type: List[Stripped]

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            continue

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            continue

        elif isinstance(our_type, intermediate.AbstractClass):
            # The abstract classes are directly dispatched by the transformer,
            # so we do not need to handle them separately.
            pass

        elif isinstance(our_type, intermediate.ConcreteClass):
            if our_type.is_implementation_specific:
                implementation_key = specific_implementations.ImplementationKey(
                    f"Jsonization/Transformer/transform_{our_type.name}.cs"
                )

                implementation = spec_impls.get(implementation_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            our_type.parsed.node,
                            f"The jsonization snippet is missing "
                            f"for the implementation-specific "
                            f"class {our_type.name}: {implementation_key}",
                        )
                    )
                    continue

                blocks.append(spec_impls[implementation_key])
            else:
                block, cls_errors = _generate_transform_for_class(cls=our_type)
                if cls_errors is not None:
                    errors.extend(cls_errors)
                else:
                    assert block is not None
                    blocks.append(block)
        else:
            assert_never(our_type)

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()
    writer.write(
        f"""\
internal class Transformer
{I}: Visitation.AbstractTransformer<Nodes.JsonObject>
{{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // internal class Transformer")

    return Stripped(writer.getvalue()), None


def _generate_serialize(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the static serializer."""
    blocks = [
        Stripped(
            "private static readonly Transformer Transformer = new Transformer();"
        ),
        Stripped(
            f"""\
/// <summary>
/// Serialize an instance of the meta-model into a JSON object.
/// </summary>
public static Nodes.JsonObject ToJsonObject(Aas.IClass that)
{{
{I}return Serialize.Transformer.Transform(that);
}}"""
        ),
    ]  # type: List[Stripped]

    for enum in symbol_table.enumerations:
        name = csharp_naming.enum_name(enum.name)
        blocks.append(
            Stripped(
                f"""\
/// <summary>
/// Serialize a literal of {name} into a JSON string.
/// </summary>
public static Nodes.JsonValue {name}ToJsonValue(Aas.{name} that)
{{
{I}string? text = Stringification.ToString(that);
{I}return Nodes.JsonValue.Create(text)
{II}?? throw new System.ArgumentException(
{III}$"Invalid {name}: {{that}}");
}}"""
            )
        )

    writer = io.StringIO()

    writer.write(
        """\
/// <summary>
/// Serialize instances of meta-model classes to JSON elements.
/// </summary>
"""
    )

    first_cls = (
        symbol_table.classes[0] if len(symbol_table.classes) > 0 else None
    )  # type: Optional[intermediate.ClassUnion]

    if first_cls is not None:
        cls_name: str

        if isinstance(first_cls, intermediate.AbstractClass):
            cls_name = csharp_naming.interface_name(first_cls.name)
        elif isinstance(first_cls, intermediate.ConcreteClass):
            cls_name = csharp_naming.class_name(first_cls.name)
        else:
            assert_never(first_cls)

        an_instance_variable = csharp_naming.variable_name(Identifier("an_instance"))

        writer.write(
            f"""\
/// <example>
/// Here is an example how to serialize an instance of {cls_name}:
/// <code>
/// var {an_instance_variable} = new Aas.{cls_name}(
///     // ... some constructor arguments ...
/// );
/// System.Text.Json.Nodes.JsonObject element = (
/// {I}Serialize.ToJsonObject(
/// {II}{an_instance_variable}));
/// </code>
/// </example>
"""
        )

    writer.write(
        """\
public static class Serialize
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // public static class Serialize")

    return Stripped(writer.getvalue())


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    namespace: csharp_common.NamespaceIdentifier,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the C# code for the general serialization.

    The ``namespace`` defines the AAS C# namespace.
    """
    errors = []  # type: List[Error]

    deserialize_impl_block, deserialize_impl_errors = _generate_deserialize_impl(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if deserialize_impl_errors is not None:
        errors.extend(deserialize_impl_errors)

    deserialize_block = _generate_deserialize(symbol_table=symbol_table)

    transformer_block, transformer_errors = _generate_transformer(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if transformer_errors is not None:
        errors.extend(transformer_errors)

    if len(errors) > 0:
        return None, errors

    assert deserialize_impl_block is not None
    assert deserialize_block is not None
    assert transformer_block is not None

    serialize_block = _generate_serialize(
        symbol_table=symbol_table,
    )

    exception_block = Stripped(
        f"""\
/// <summary>
/// Represent a critical error during the deserialization.
/// </summary>
public class Exception : System.Exception
{{
{I}public readonly string Path;
{I}public readonly string Cause;
{I}public Exception(string path, string cause)
{II}: base($"{{cause}} at: {{path}}")
{I}{{
{II}Path = path;
{II}Cause = cause;
{I}}}
}}"""
    )

    jsonization_blocks = [
        deserialize_impl_block,
        exception_block,
        deserialize_block,
        transformer_block,
        serialize_block,
    ]  # type: List[Stripped]

    jsonization_writer = io.StringIO()
    jsonization_writer.write(
        f"""\
namespace {namespace}
{{
{I}/// <summary>
{I}/// Provide de/serialization of meta-model classes to/from JSON.
{I}/// </summary>
{I}/// <remarks>
{I}/// We can not use one-pass deserialization for JSON since the object
{I}/// properties do not have fixed order, and hence we can not read
{I}/// <c>modelType</c> property ahead of the remaining properties.
{I}/// </remarks>
{I}public static class Jsonization
{I}{{
"""
    )

    for i, deserialize_block in enumerate(jsonization_blocks):
        if i > 0:
            jsonization_writer.write("\n\n")

        jsonization_writer.write(textwrap.indent(deserialize_block, II))

    jsonization_writer.write(f"\n{I}}}  // public static class Jsonization")
    jsonization_writer.write(f"\n}}  // namespace {namespace}")

    using_directives = []  # type: List[Stripped]
    using_directives.extend(
        csharp_common.generate_using_aas_directive_if_necessary(namespace)
    )

    using_directives.append(
        Stripped(
            """\
using CodeAnalysis = System.Diagnostics.CodeAnalysis;
using Nodes = System.Text.Json.Nodes;

using System.Collections.Generic;  // can't alias"""
        )
    )

    # pylint: disable=line-too-long
    blocks = [
        csharp_common.WARNING,
        Stripped("\n".join(using_directives)),
        Stripped(jsonization_writer.getvalue()),
        csharp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        assert not block.startswith("\n")
        assert not block.endswith("\n")
        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None



"""Provide functions shared among different TypeScript code generation modules."""
import io
import math
from typing import List, Tuple, Optional, Union

from icontract import require

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, assert_never, Identifier
from aas_core_codegen.typescript import naming as typescript_naming


def boolean_literal(value: bool) -> Stripped:
    """Generate the boolean literal corresponding to the ``value``."""
    return Stripped("true") if value else Stripped("false")


def representable_as_number(value: int) -> bool:
    """Check that the ``value`` can be represented as a double-precision float."""
    return float(value) == value


@require(lambda value: not isinstance(value, int) or representable_as_number(value))
def numeric_literal(value: Union[int, float]) -> Stripped:
    """Generate the numeric literal corresponding to the ``value``."""
    if math.isnan(value):
        return Stripped("NaN")
    if value == math.inf:
        return Stripped("Infinity")
    elif value == -math.inf:
        return Stripped("-Infinity")
    else:
        return Stripped(str(value))


# See: https://262.ecma-international.org/5.1/#sec-7.8.4
_BASE_ESCAPING_IN_TYPESCRIPT = {
    "\\": "\\\\",
    "\b": "\\b",
    "\t": "\\t",
    "\n": "\\n",
    "\v": "\\v",
    "\f": "\\f",
    "\r": "\\r",
}


def string_literal(
    text: str,
    without_enclosing: bool = False,
    in_backticks: bool = False,
) -> Stripped:
    """
    Generate a string literal from the ``text``.

    If ``without_enclosing`` is set, the enclosing quotes are omitted.

    If ``in_backticks`` is set, the enclosing quotes are assumed to be backticks
    and the escaping is performed according to:
    https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Template_literals
    """
    escaped_chars = []  # type: List[str]

    if len(text) > 0:
        iterator = iter(text)
        current_char = next(iterator, None)  # type: Optional[str]
        assert (
            current_char is not None
        ), "If `text` is non-empty, we have to observe at least a single character"

        while current_char is not None:
            next_char = next(iterator, None)

            escaped_char = _BASE_ESCAPING_IN_TYPESCRIPT.get(current_char, None)
            if escaped_char is not None:
                escaped_chars.append(escaped_char)
            else:
                if not in_backticks:
                    if current_char == '"':
                        escaped_chars.append('\\"')
                    else:
                        escaped_chars.append(current_char)
                else:
                    if current_char == "`":
                        escaped_chars.append("\\`")
                    elif (
                        current_char == "$"
                        and next_char is not None
                        and next_char == "{"
                    ):
                        escaped_chars.append("\\$")
                    else:
                        escaped_chars.append(current_char)

            current_char = next_char

    escaped = "".join(escaped_chars)

    if without_enclosing:
        return Stripped(escaped)
    else:
        if not in_backticks:
            return Stripped(f'"{escaped}"')
        else:
            return Stripped(f"`{escaped}`")


INDENT = "  "
INDENT2 = INDENT * 2


def bytes_literal(value: bytes) -> Tuple[Stripped, bool]:
    """
    Generate an expression representing the ``value``.

    If there are more than 8 bytes, a multi-line expression is returned.

    :param value: to be represented
    :return: (TypeScript expression, is multi-line)
    """
    if len(value) == 0:
        return Stripped("new Uint8Array()"), False

    writer = io.StringIO()

    if len(value) <= 8:
        items_joined = ", ".join(f"0x{byte:02x}" for byte in value)
        return Stripped(f"new Uint8Array([{items_joined}])"), False
    else:
        writer.write(
            f"""\
new Uint8Array(
{INDENT}["""
        )

        for start in range(0, len(value), 8):
            if start == 0:
                writer.write(f"\n{INDENT2}")
            else:
                writer.write(f",\n{INDENT2}")

            end = min(start + 8, len(value))

            assert start < end

            for i, byte in enumerate(value[start:end]):
                if i > 0:
                    writer.write(", ")

                writer.write(f"0x{byte:02x}")

        writer.write(f"\n{INDENT}]\n)")

        return Stripped(writer.getvalue()), True


def needs_escaping(text: str, in_backticks: bool = False) -> bool:
    """
    Check whether the ``text`` contains a character that needs escaping.

    If ``in_backticks`` is set, it checks that the ``text`` needs not be escaped if
    enclosed in backticks (instead of double quotes).
    """
    prev_character = None  # type: Optional[str]
    for character in text:
        if character in _BASE_ESCAPING_IN_TYPESCRIPT:
            return True

        if not in_backticks:
            if character == '"':
                return True
        else:
            if character == "`":
                return True

            if (
                prev_character is not None
                and prev_character == "$"
                and character == "{"
            ):
                return True

        prev_character = character

    return False


PRIMITIVE_TYPE_MAP = {
    intermediate.PrimitiveType.BOOL: Stripped("boolean"),
    intermediate.PrimitiveType.INT: Stripped("number"),
    intermediate.PrimitiveType.FLOAT: Stripped("number"),
    intermediate.PrimitiveType.STR: Stripped("string"),
    intermediate.PrimitiveType.BYTEARRAY: Stripped("Uint8Array"),
}


def _assert_all_primitive_types_are_mapped() -> None:
    """Assert that we have explicitly mapped all the primitive types to C#."""
    all_primitive_literals = set(literal.value for literal in PRIMITIVE_TYPE_MAP)

    mapped_primitive_literals = set(
        literal.value for literal in intermediate.PrimitiveType
    )

    all_diff = all_primitive_literals.difference(mapped_primitive_literals)
    mapped_diff = mapped_primitive_literals.difference(all_primitive_literals)

    messages = []  # type: List[str]
    if len(mapped_diff) > 0:
        messages.append(
            f"More primitive maps are mapped than there were defined "
            f"in the ``intermediate._types``: {sorted(mapped_diff)}"
        )

    if len(all_diff) > 0:
        messages.append(
            f"One or more primitive types in the ``intermediate._types`` were not "
            f"mapped in PRIMITIVE_TYPE_MAP: {sorted(all_diff)}"
        )

    if len(messages) > 0:
        raise AssertionError("\n\n".join(messages))


_assert_all_primitive_types_are_mapped()


def generate_type(
    type_annotation: intermediate.TypeAnnotationUnion,
    types_module: Optional[Stripped] = None,
) -> Stripped:
    """
    Generate the type for the given type annotation.

    If ``types_module`` is specified, it is used as prefix for the composite types.
    """
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return PRIMITIVE_TYPE_MAP[type_annotation.a_type]

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type

        name: Identifier

        if isinstance(our_type, intermediate.Enumeration):
            name = typescript_naming.enum_name(type_annotation.our_type.name)

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return PRIMITIVE_TYPE_MAP[our_type.constrainee]

        elif isinstance(our_type, intermediate.ConcreteClass):
            name = typescript_naming.class_name(type_annotation.our_type.name)

        elif isinstance(our_type, intermediate.AbstractClass):
            name = typescript_naming.interface_name(type_annotation.our_type.name)

        else:
            assert_never(our_type)

        return Stripped(name if types_module is None else f"{types_module}.{name}")

    elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
        item_type = generate_type(
            type_annotation=type_annotation.items, types_module=types_module
        )

        return Stripped(f"Array<{item_type}>")

    elif isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        value = generate_type(
            type_annotation=type_annotation.value, types_module=types_module
        )

        return Stripped(f"{value} | null")

    else:
        assert_never(type_annotation)

    raise AssertionError("Should not have gotten here")


INDENT3 = INDENT * 3
INDENT4 = INDENT * 4
INDENT5 = INDENT * 5
INDENT6 = INDENT * 6

WARNING = Stripped(
    """\
// This code has been automatically generated by aas-core-codegen.
// Do NOT edit or append."""
)


class GeneratorForLoopVariables:
    """
    Generate a unique variable name based on ``item`` stem.

    >>> generator = GeneratorForLoopVariables()

    >>> next(generator)
    'anItem'

    >>> next(generator)
    'anotherItem'

    >>> next(generator)
    'yetAnotherItem'

    >>> next(generator)
    'yetYetAnotherItem'
    """

    def __init__(self) -> None:
        """Initialize with the zero counter."""
        self.counter = 0

    def __next__(self) -> Identifier:
        """Generate the next variable name."""
        if self.counter == 0:
            result = Identifier("anItem")
        elif self.counter == 1:
            result = Identifier("anotherItem")
        elif self.counter == 2:
            result = Identifier("yetAnotherItem")
        else:
            result = Identifier("yet" + ("Yet" * (self.counter - 2)) + "AnotherItem")

        self.counter += 1

        return result


#: Name of the module where all the types are defined
TYPES_MODULE = Identifier("types")

#: Name of the module where all the constants are defined
CONSTANTS_MODULE = Identifier("constants")

#: Name of the module where all the verification logic resides
VERIFICATION_MODULE = Identifier("verification")
